import warnings

import numpy as np

__all__ = ["timtraceline", "timtracelines"]

_future_warning_metadata = (
    "In a future version traces will be returned as a dictionary containing "
    "metadata together with the trace. To already get the new behavior use "
    "metadata=True."
)


def timtraceline(ml, xstart, ystart, zstart, hstepmax, vstepfrac=0.2, tmax=1e12,
                 nstepmax=100, win=[-1e30, 1e30, -1e30, 1e30], silent=False, 
                 returnlayers=False, *, metadata=False):
    verbose = False  # used for debugging
    if not metadata:
        warnings.warn(_future_warning_metadata, FutureWarning, stacklevel=2)
    # treating aquifer layers and leaky layers the same way
    xw1, xw2, yw1, yw2 = win
    terminate = False
    message = "no message"
    eps = 1e-10  # used to place point just above or below aquifer top or bottom
    direction = np.sign(hstepmax)  # negative means backwards
    hstepmax = np.abs(hstepmax)
    aq = ml.aq.find_aquifer_data(xstart, ystart)
    if zstart > aq.z[0] or zstart < aq.z[-1]:
        terminate = True
        message = "starting z value not inside aquifer"
    layer, ltype, modellayer = aq.findlayer(zstart)
    # slightly alter starting location not to get stuck in surpring points
    # starting at time 0
    xyzt = [np.array([xstart * (1 + eps), ystart * (1 + eps), zstart, 0])]  
    layerlist = []  # to keep track of layers for plotting with colors
    for _ in range(nstepmax):
        if terminate:
            break
        x0, y0, z0, t0 = xyzt[-1]
        aq = ml.aq.find_aquifer_data(x0, y0)  # find new aquifer
        layer, ltype, modellayer = aq.findlayer(z0)
        layerlist.append(modellayer)
        v0 = ml.velocomp(x0, y0, z0, aq, [layer, ltype]) * direction
        if verbose:
            print('xyz, layer', x0, y0, z0, layer)
            print('v0, layer, ltype', v0, layer, ltype)
        vx, vy, vz = v0
        if ltype == "l":  # in leaky layer
            if vz > 0:  # upward through leaky layer
                if modellayer == 0:  # steps out of the top
                    z1 = aq.z[modellayer]
                    terminate = True
                else:
                    modellayer -= 1
                    # just above new bottom
                    z1 = aq.z[modellayer + 1] + eps * aq.Hlayer[modellayer]
            elif vz < 0:
                if modellayer == aq.nlayers - 1:  # steps out of bottom
                    z1 = aq.z[modellayer + 1]
                    terminate = True
                else:
                    modellayer += 1
                    # just below new top
                    z1 = aq.z[modellayer] - eps * aq.Hlayer[modellayer]
            else:
                message = "at point of zero leakage in leaky layer"
                terminate = True
                break
            t1 = t0 + abs((z1 - z0) / vz)
            xyztnew = [np.array([x0, y0, z1, t1])]
        else:  # in aquifer layer
            vh = np.sqrt(vx ** 2 + vy ** 2)
            if vz > 0:  # flows upward
                if aq.z[modellayer] - z0 < vstepfrac * aq.Haq[layer]:
                    # just below top
                    z1 = aq.z[modellayer] - eps * aq.Hlayer[modellayer]
                else:
                    z1 = z0 + vstepfrac * aq.Haq[layer]
                tvstep = (z1 - z0) / vz
            elif vz < 0:
                if z0 - aq.z[modellayer + 1] < vstepfrac * aq.Haq[layer]:
                    # just above bot
                    z1 = aq.z[modellayer + 1] + eps * aq.Hlayer[modellayer]
                else:
                    z1 = z0 - vstepfrac * aq.Haq[layer]
                tvstep = (z0 - z1) / abs(vz)
            else:  # vz=0
                tvstep = np.inf
                z1 = z0
            if tvstep == np.inf and vh == 0:  # this should never happen anymore
                message = "at point of zero velocity"
                terminate = True
                break
            if vh * tvstep > hstepmax:  
                # max horizonal step smaller than max vertical step
                thstep = hstepmax / vh
                z1 = z0 + thstep * vz
            else:
                thstep = tvstep
                # z1 is already computed
            x1 = x0 + thstep * vx
            y1 = y0 + thstep * vy
            t1 = t0 + thstep
            xyzt1 = np.array([x1, y1, z1, t1])
            # check if point needs to be changed
            correction = True
            for e in aq.elementlist:
                changed, terminate, xyztnew, changemessage = e.changetrace(
                    xyzt[-1], xyzt1, aq, layer, ltype, modellayer, 
                    direction, hstepmax)
                if changed or terminate:
                    correction = False
                    if changemessage:
                        message = changemessage
                    break
            if correction:  # correction step
                vx, vy, vz = 0.5 * (v0 + direction * \
                                    ml.velocomp(x1, y1, z1, aq, [layer, ltype]))
                if verbose:
                    print('xyz1, layer', x1, y1, z1, layer)
                    print('correction vx, vy, vz', vx, vy, vz)
                vh = np.sqrt(vx ** 2 + vy ** 2)
                if vz > 0:  # flows upward
                    tvstep = min(aq.z[modellayer] - z0, 
                                 vstepfrac * aq.Haq[layer]) / vz
                elif vz < 0:
                    tvstep = min(z0 - aq.z[modellayer + 1], 
                                 vstepfrac * aq.Haq[layer]) / abs(vz)
                else:  # vz=0
                    tvstep = np.inf
                if vh * tvstep > hstepmax:
                    # max horizonal step smaller than vertical step
                    thstep = hstepmax / vh
                    x1 = x0 + thstep * vx
                    y1 = y0 + thstep * vy
                    z1 = z0 + thstep * vz
                else:
                    thstep = tvstep
                    x1 = x0 + thstep * vx
                    y1 = y0 + thstep * vy
                    if vz > 0:  # flows upward
                        if aq.z[modellayer] - z0 < vstepfrac * aq.Haq[layer]:
                            if modellayer == 0:  # steps out of the top
                                z1 = aq.z[modellayer]
                                terminate = True
                                message = "flowed out of top"
                            else:
                                modellayer -= 1
                                # just above new bottom
                                z1 = aq.z[modellayer + 1] + \
                                     eps * aq.Hlayer[modellayer]
                        else:
                            z1 = z0 + thstep * vz
                    else:
                        if z0 - aq.z[modellayer + 1] < vstepfrac * aq.Haq[layer]:
                            if modellayer == aq.nlayers - 1:  # steps out of bottom
                                z1 = aq.z[modellayer + 1]
                                terminate = True
                                message = "flowed out of bottom"
                            else:
                                modellayer += 1
                                # just below new top
                                z1 = aq.z[modellayer] - \
                                     eps * aq.Hlayer[modellayer]
                        else:
                            z1 = z0 + thstep * vz
                    if not terminate:
                        layer = aq.layernumber[modellayer]
                        ltype = aq.ltype[modellayer]
                t1 = t0 + thstep
                xyztnew = [np.array([x1, y1, z1, t1])]
                # check again if point needs to be changed
                for e in aq.elementlist:
                    changed, terminate, xyztchanged, changemessage = \
                        e.changetrace(xyzt[-1],  xyztnew[0], aq, layer, ltype,
                                      modellayer, direction, hstepmax)
                    if changed or terminate:
                        xyztnew = xyztchanged
                        if changemessage:
                            message = changemessage
                        break
        # check if outside window
        x1, y1, z1, t1 = xyztnew[0]
        frac = -1  # used to check later whether something changed
        if x1 < xw1:
            frac = abs((x0 - xw1) / (x1 - x0))
            x1, y1, z1, t1 = xyzt[-1] + frac * (xyztnew[0] - xyzt[-1])
            message = "reached window boundary"
        if x1 > xw2:
            frac = abs((x0 - xw2) / (x1 - x0))
            x1, y1, z1, t1 = xyzt[-1] + frac * (xyztnew[0] - xyzt[-1])
            message = "reached window boundary"
        if y1 < yw1:
            frac = abs((y0 - yw1) / (y1 - y0))
            x1, y1, z1, t1 = xyzt[-1] + frac * (xyztnew[0] - xyzt[-1])
            message = "reached window boundary"
        if y1 > yw2:
            frac = abs((y0 - yw2) / (y1 - y0))
            x1, y1, z1, t1 = xyzt[-1] + frac * (xyztnew[0] - xyzt[-1])
            message = "reached window boundary"
        if t1 > tmax:
            frac = abs((tmax - t0) / (t1 - t0))
            x1, y1, z1, t1 = xyzt[-1] + frac * (xyztnew[0] - xyzt[-1])
            message = "reached tmax"
        if frac > 0:  # at least one of the above 5 ifs was true
            terminate = True
            xyztnew = [np.array([x1, y1, z1, t1])]
        xyzt.extend(xyztnew)
        if len(xyztnew) == 2:
            layerlist.append(modellayer)
        elif len(xyztnew) > 3:
            print("len(xyztnew > 3 !")
            print(xyztnew)
    else:
        message = "reached nstepmax iterations"
    if not silent:
        print(message)
    if metadata:
        result = {"trace": np.array(xyzt), "message": message, 
                  "complete": terminate}
        if returnlayers:
            result["layers"] = layerlist
    else:
        if returnlayers:
            result = np.array(xyzt), layerlist
        else:
            result = np.array(xyzt)
    return result


def timtracelines(ml, xstart, ystart, zstart, hstepmax, vstepfrac=0.2, 
                  tmax=1e12, nstepmax=100, silent=".", 
                  win=[-1e30, 1e30, -1e30, 1e30], *, metadata=False):
    xyztlist = []
    for x, y, z in zip(xstart, ystart, zstart):
        xyztlist.append(
            timtraceline(ml, x, y, z, hstepmax=hstepmax, vstepfrac=vstepfrac,
                         tmax=tmax, nstepmax=nstepmax, silent=silent, win=win,
                         metadata=metadata)
        )
        if silent == ".":
            print(".", end="", flush=True)
    if silent == ".":
        print("")
    return xyztlist


def crossline(xa, ya, xb, yb, z1, z2):
    eps = 1e-8
    za = xa + ya * 1j
    zb = xb + yb * 1j
    Za = (2 * za - (z1 + z2)) / (z2 - z1)
    Zb = (2 * zb - (z1 + z2)) / (z2 - z1)
    if Za.imag * Zb.imag < 0:
        Xa, Ya = Za.real, Za.imag
        Xb, Yb = Zb.real, Zb.imag
        X = Xa - Ya * (Xb - Xa) / (Yb - Ya)
        if abs(X) <= 1:
            Z = X + eps * np.sign(Yb) * 1j
            z = 0.5 * ((z2 - z1) * Z + z1 + z2)
            return True, z.real, z.imag
    return False



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import LineCollection

from .trace import timtraceline, timtracelines

plt.rcParams["contour.negative_linestyle"] = "solid"


class PlotTim:
    def plot(self, win=None, newfig=True, figsize=None, orientation="hor",
             topfigfrac=0.8, layer=None):
        """Plot layout

        Parameters
        ----------

        win : list or tuple
            [x1, x2, y1, y2]

        """

        if newfig:
            plt.figure(figsize=figsize)
            ax1 = None
            ax2 = None
            if orientation == "both":
                ax1 = plt.axes([0.125, 0.18 + (1 - topfigfrac) * 0.7,
                                (0.9 - 0.125), topfigfrac * 0.7])
                ax2 = plt.axes([0.125, 0.11, (0.9 - 0.125), 
                                (1 - topfigfrac) * 0.7], sharex=ax1)
            elif orientation[:3] == "hor":
                ax1 = plt.subplot()
            elif orientation[:3] == "ver":
                ax2 = plt.subplot()
        else:
            if orientation == "both":
                fig = plt.gcf()
                ax1 = fig.axes[0]
                ax2 = fig.axes[1]
            elif orientation[:3] == "hor":
                fig = plt.gcf()
                ax1 = fig.axes[0]
                ax2 = None
            elif orientation[:3] == "ver":
                fig = plt.gcf()
                ax1 = None
                ax2 = fig.axes[0]
        if ax1 is not None:
            plt.sca(ax1)
            for e in self.elementlist:
                e.plot(layer=layer)
            if orientation[:3] == "hor":
                plt.axis("scaled")
            elif orientation == "both":
                plt.axis("equal")  # cannot be 'scaled' when sharing axes
            if win is not None:
                plt.axis(win)
        if ax2 is not None:
            plt.sca(ax2)
            for i in range(self.aq.nlayers):
                if self.aq.ltype[i] == "l":
                    plt.axhspan(ymin=self.aq.z[i + 1], ymax=self.aq.z[i],
                                color=[0.8, 0.8, 0.8])
            for i in range(1, self.aq.nlayers):
                if self.aq.ltype[i] == "a" and self.aq.ltype[i - 1] == "a":
                    plt.axhspan(ymin=self.aq.z[i], ymax=self.aq.z[i], 
                                color=[0.8, 0.8, 0.8])

    def contour(self, win, ngr=20, layers=0, levels=20, layout=True,
                labels=True, decimals=0, color=None, newfig=True,
                figsize=None, legend=True):
        """Contour plot

        Parameters
        ----------

        win : list or tuple
            [x1, x2, y1, y2]
        ngr : scalar, tuple or list
            if scalar: number of grid points in x and y direction
            if tuple or list: nx, ny, number of grid points in x and y
            directions
        layers : integer, list or array
            layers for which grid is returned
        levels : integer or array (default 20)
            levels that are contoured
        layout : boolean (default True)
            plot layout of elements
        labels : boolean (default True)
            print labels along contours
        decimals : integer (default 0)
            number of decimals of labels along contours
        color : str or list of strings
            color of contour lines
        newfig : boolean (default True)
            create new figure
        figsize : tuple of 2 values (default is mpl default)
            size of figure
        legend : list or boolean (default True)
            add legend to figure
            if list of strings: use strings as names in legend

        """

        x1, x2, y1, y2 = win
        if np.isscalar(ngr):
            nx = ny = ngr
        else:
            nx, ny = ngr
        layers = np.atleast_1d(layers)
        xg = np.linspace(x1, x2, nx)
        yg = np.linspace(y1, y2, ny)
        h = self.headgrid(xg, yg, layers)
        if newfig:
            plt.figure(figsize=figsize)
        # color
        if color is None:
            c = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        elif type(color) is str:
            c = len(layers) * [color]
        elif type(color) is list:
            c = color
        if len(c) < len(layers):
            n = np.ceil(self.aq.naq / len(c))
            c = n * c
        # contour
        cscollectionlist = []
        for i in range(len(layers)):
            cs = plt.contour(xg, yg, h[i], levels, colors=c[i])
            cscollectionlist.append(cs.collections[0])
            if labels:
                fmt = "%1." + str(decimals) + "f"
                plt.clabel(cs, fmt=fmt)
        if type(legend) is list:
            plt.legend(cscollectionlist, legend)
        elif legend:
            legendlist = ["layer " + str(i) for i in layers]
            plt.legend(cscollectionlist, legendlist)
        plt.axis("scaled")
        if layout:
            self.plot(win=[x1, x2, y1, y2], newfig=False, layer=layers)
        # plt.show()

    def vcontour(self, win, n, levels, labels=False, decimals=0, color=None,
                 vinterp=True, nudge=1e-6, newfig=True, figsize=None,
                 layout=True):
        """Vertical contour
        """
        x1, x2, y1, y2 = win
        h = self.headalongline(
            np.linspace(x1 + nudge, x2 - nudge, n),
            np.linspace(y1 + nudge, y2 - nudge, n),
        )
        L = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        xg = np.linspace(0, L, n)
        if vinterp:
            zg = 0.5 * (self.aq.zaqbot + self.aq.zaqtop)
            zg = np.hstack((self.aq.zaqtop[0], zg, self.aq.zaqbot[-1]))
            h = np.vstack((h[0], h, h[-1]))
        else:
            zg = np.empty(2 * self.aq.naq)
            for i in range(self.aq.naq):
                zg[2 * i] = self.aq.zaqtop[i]
                zg[2 * i + 1] = self.aq.zaqbot[i]
            h = np.repeat(h, 2, 0)
        if newfig:
            plt.figure(figsize=figsize)
        cs = plt.contour(xg, zg, h, levels, colors=color)
        if labels:
            fmt = "%1." + str(decimals) + "f"
            plt.clabel(cs, fmt=fmt)
        if layout:
            self.plot(win=[x1, x2, y1, y2], orientation="ver", newfig=False)

    def tracelines(self, xstart, ystart, zstart, hstepmax, vstepfrac=0.2,
                   tmax=1e12, nstepmax=100, silent=".", color=None,
                   orientation="hor", win=[-1e30, 1e30, -1e30, 1e30],
                   newfig=False, figsize=None, *, return_traces=False,
                   metadata=False):
        """Draw trace lines
        """
        if color is None:
            c = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        elif type(color) is str:
            c = self.aq.naq * [color]
        elif type(color) is list:
            c = color
        if len(c) < self.aq.naq:
            n = int(np.ceil(self.aq.naq / len(c)))
            c = n * c
        fig = plt.gcf()
        assert (len(fig.axes) > 0), \
            "Error: Need to specify axes in figure before invoking tracelines"
        ax1 = None
        ax2 = None
        if orientation == "both":
            ax1 = fig.axes[0]
            ax2 = fig.axes[1]
        elif orientation[:3] == "hor":
            ax1 = fig.axes[0]
        elif orientation[:3] == "ver":
            ax2 = fig.axes[1]
        if return_traces:
            traces = []
        else:
            metadata = True  # suppress future warning from timtraceline
        for i in range(len(xstart)):
            trace = timtraceline(
                self,
                xstart[i],
                ystart[i],
                zstart[i],
                hstepmax=hstepmax,
                vstepfrac=vstepfrac,
                tmax=tmax,
                nstepmax=nstepmax,
                silent=silent,
                win=win,
                returnlayers=True,
                metadata=metadata,
            )
            if return_traces:
                traces.append(trace)
            if metadata:
                xyzt, layerlist = trace["trace"], trace["layers"]
            else:
                xyzt, layerlist = trace
            if silent == ".":
                print(".", end="", flush=True)
            if ax1 is not None:
                # plt.axes(ax1)
                color = [
                    c[self.aq.layernumber[i]] if self.aq.ltype[i] == "a" else "k"
                    for i in layerlist
                ]
                points = np.array([xyzt[:, 0], xyzt[:, 1]]).T.reshape(-1, 1, 2)
                segments = np.concatenate([points[:-1], points[1:]], axis=1)
                lc = LineCollection(segments, colors=color)
                ax1.add_collection(lc)
                # ax1.plot(xyzt[:, 0], xyzt[:, 1], color=color)
            if ax2 is not None:
                color = [
                    c[self.aq.layernumber[i]] if self.aq.ltype[i] == "a" else "k"
                    for i in layerlist
                ]
                points = np.array([xyzt[:, 0], xyzt[:, 2]]).T.reshape(-1, 1, 2)
                segments = np.concatenate([points[:-1], points[1:]], axis=1)
                lc = LineCollection(segments, colors=color)
                ax2.add_collection(lc)
                ax2.set_ylim(self.aq.z[-1], self.aq.z[0])
        if silent == ".":
            print("")  # Print the final newline after the dots
        if return_traces:
            return traces

    def vcontoursf1D(self, x1, x2, nx, levels, labels=False, decimals=0,
                     color=None, nudge=1e-6, newfig=True, figsize=None,
                     layout=True, ax=None):
        """
        Vertical contour for 1D model

        """
        naq = self.aq.naq
        xflow = np.linspace(x1 + nudge, x2 - nudge, nx)
        Qx = np.empty((naq, nx))
        for i in range(nx):
            Qx[:, i], _ = self.disvec(xflow[i], 0)
        zflow = np.empty(2 * naq)
        for i in range(self.aq.naq):
            zflow[2 * i] = self.aq.zaqtop[i]
            zflow[2 * i + 1] = self.aq.zaqbot[i]
        Qx = Qx[::-1]  # set upside down
        Qxgrid = np.empty((2 * naq, nx))
        Qxgrid[0] = 0
        for i in range(naq - 1):
            Qxgrid[2 * i + 1] = Qxgrid[2 * i] - Qx[i]
            Qxgrid[2 * i + 2] = Qxgrid[2 * i + 1]
        Qxgrid[-1] = Qxgrid[-2] - Qx[-1]
        Qxgrid = Qxgrid[::-1]  # index 0 at top
        if newfig:
            _, ax = plt.subplots(1, 1, figsize=figsize)
        else:
            ax = ax
        cs = ax.contour(xflow, zflow, Qxgrid, levels, colors=color)
        if labels:
            fmt = "%1." + str(decimals) + "f"
            plt.clabel(cs, fmt=fmt)
        return ax
        # if layout:
        #    self.plot(win=[x1, x2, y1, y2], orientation='ver', newfig=False)



"""
Model classes

"""

import numpy as np
import sys
import inspect  # Used for storing the input
from .aquifer import Aquifer
from .aquifer_parameters import param_maq, param_3d
from .constant import ConstantStar
from .util import PlotTim
import multiprocessing as mp

__all__ = ['Model', 'ModelMaq', 'Model3D']

class Model(PlotTim):
    """
    Model Class to create a model object consisting of an arbitrary
    sequence of aquifer layers and leaky layers.
    Use ModelMaq for regular sequence of aquifers and leaky layers.
    Use Model3D for multi-layer model of a single aquifer
    
    Parameters
    ----------
    kaq : array
        hydraulic conductivity of each aquifer from the top down
    z : array
        elevation tops and bottoms of all layers
        layers may have zero thickness
    c : array
        resistance between two consecutive aquifer layers
        if ltype[0]='a': length is number of aquifers - 1
        if ltype[0]='l': length is number of aquifers
    npor : array
        porosity of all layers from the top down
    ltype : array of characters
        array indicating for each layer whether it is
        'a' aquifer layer
        'l' leaky layer
    
    """
    
    def __init__(self, kaq, c, z, npor, ltype, f2py=False):
        # All input variables are numpy arrays
        # That should be checked outside this function
        self.elementlist = []
        self.elementdict = {}  # only elements that have a label
        self.aq = Aquifer(self, kaq, c, z, npor, ltype)
        self.modelname = 'ml'  # Used for writing out input
        self.f2py = False
        if f2py:
            try:
                from .src import besselaesnew
                self.f2py = True
            except:
                print('FORTRAN extension not found while f2py=True')
                print('Using Numba instead')

    def initialize(self):
        # remove inhomogeneity elements (they are added again)
        self.elementlist = [e for e in self.elementlist if not e.inhomelement]
        self.aq.initialize()
        for e in self.elementlist:
            e.initialize()

    def add_element(self, e):
        self.elementlist.append(e)
        if e.label is not None: self.elementdict[e.label] = e

    def remove_element(self, e):
        """Remove element `e` from model
        """
        
        if e.label is not None: self.elementdict.pop(e.label)
        self.elementlist.remove(e)

    def storeinput(self, frame):
        self.inputargs, _, _, self.inputvalues = inspect.getargvalues(frame)

    def potential(self, x, y, aq=None):
        if aq is None: aq = self.aq.find_aquifer_data(x, y)
        pot = np.zeros(aq.naq)
        for e in aq.elementlist:
            pot += e.potential(x, y, aq)
        rv = np.sum(pot * aq.eigvec, 1)
        if aq.ltype[0] == 'l':
            # potential for head above leaky layer
            rv += aq.constantstar.potstar
        return rv

    def disvec(self, x, y, aq=None):
        """Discharge vector at `x`, `y`
        
        Returns
        -------
        
        qxqy : array size (2, naq)
            first row is Qx in each aquifer layer, second row is Qy
        """
        
        if aq is None: aq = self.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, aq.naq))
        for e in aq.elementlist:
            rv += e.disvec(x, y, aq)
        rv = np.sum(rv[:, np.newaxis, :] * aq.eigvec, 2)
        return rv
    
    def qztop(self, x, y, aq=None):
        if aq is None: aq = self.aq.find_aquifer_data(x, y)
        rv = 0.0
        if aq.ltype[0] == 'a':  # otherwise recharge cannot be added
            for e in aq.elementlist:
                rv += e.qztop(x, y)
        return rv

    def head(self, x, y, layers=None, aq=None):
        """Head at `x`, `y`
        
        Returns
        -------
        
        h : array length `naq` or `len(layers)`
            head in all `layers` (if not `None`), 
            or all layers of aquifer (otherwise)
        """
        
        if aq is None: aq = self.aq.find_aquifer_data(x, y)
        rv = self.potential(x, y, aq) / aq.T
        if layers is None:
            return rv
        else:
            return rv[layers]

    def headgrid(self, xg, yg, layers=None, printrow=False):
        """Grid of heads
        
        Parameters
        ----------
        xg : array
            x values of grid
        yg : array
            y values of grid
        layers : integer, list or array, optional
            layers for which grid is returned
        printrow : boolean, optional
            prints dot to screen for each row of grid if set to `True`
        
        Returns
        -------
        h : array size `nlayers, ny, nx`
        
        See also
        --------
        
        :func:`~timml.model.Model.headgrid2`

        """
        
        nx, ny = len(xg), len(yg)
        if layers is None:
            Nlayers = self.aq.find_aquifer_data(xg[0], yg[0]).naq
        else:
            Nlayers = len(np.atleast_1d(layers))
        h = np.empty((Nlayers, ny, nx))
        for j in range(ny):
            if printrow:
                print('.', end='', flush=True)
            for i in range(nx):
                h[:, j, i] = self.head(xg[i], yg[j], layers)
        if printrow:
            print('', flush=True)
        return h

    def headgrid2(self, x1, x2, nx, y1, y2, ny, layers=None, printrow=False):
        """Grid of heads
        
        Parameters
        ----------
        x1, x2, nx : 
            x values are generated as linspace(x1, x2, nx)
        y1, y2, ny : 
            y values are generated as linspace(y1, y2, ny)
        layers : integer, list or array, optional
            layers for which grid is returned
        printrow : boolean, optional
            prints dot to screen for each row of grid if set to `True`
        
        Returns
        -------
        h : array size `nlayers, ny, nx`
        
        See also
        --------
        
        :func:`~timml.model.Model.headgrid`
        
        """
        
        xg, yg = np.linspace(x1, x2, nx), np.linspace(y1, y2, ny)
        return self.headgrid(xg, yg, layers=layers, printrow=printrow)

    def headalongline(self, x, y, layers=None):
        """Head along line or curve
        
        Parameters
        ----------
        x : array
            x values of line
        y : array
            y values of line
        layers : integer, list or array, optional
            layers for which grid is returned
        
        Returns
        -------
        h : array size `nlayers, nx`

        """
        
        xg, yg = np.atleast_1d(x), np.atleast_1d(y)
        if layers is None:
            Nlayers = self.aq.find_aquifer_data(xg[0], yg[0]).naq
        else:
            Nlayers = len(np.atleast_1d(layers))
        nx = len(xg)
        if len(yg) == 1:
            yg = yg * np.ones(nx)
        h = np.zeros((Nlayers, nx))
        for i in range(nx):
            h[:, i] = self.head(xg[i], yg[i], layers)
        return h
    
    def disvecalongline(self, x, y, layers=None):
        '''Returns Qx[Nlayers,len(x)], Qy[Nlayers,len(x)]
        Assumes same number of layers for each x and y
        layers may be None or list of layers for which head is computed'''
        xg, yg = np.atleast_1d(x), np.atleast_1d(y)
        if layers is None:
            nlayers = self.aq.find_aquifer_data(xg[0], yg[0]).naq
        else:
            nlayers = len(np.atleast_1d(layers))
        nx = len(xg)
        if len(yg) == 1:
            yg = yg * np.ones(nx)
        Qx = np.zeros((nlayers, nx))
        Qy = np.zeros((nlayers, nx))
        for i in range(nx):
            Qx[:, i], Qy[:, 1] = self.disvec(xg[i], yg[i], layers)
        return Qx, Qy
    
#    def disvec_direction(self, s, x1, y1, cdirection):
#        pass
#    
#    def discharge_across_line(self, x1, y1, x2,  y2, layers=None):
#        if layers is None:
#            nlayers = self.aq.find_aquifer_data(x1, y1).naq
#        else:
#            nlayers = len(np.atleast_1d(layers))
#        z1 = x1 + y1 * 1j
#        z2 = x2 + y2 * 1j
#        normvec = (z2 - z1) / np.abs(z2 - z1) * np.exp(-np.pi * 1j / 2)
#        disvec = self.disvec(xg[i], yg[i], layers)
    
    def velocity(self, x, y, z):
        return self.velocomp(x, y, z)
    
    def velocomp(self, x, y, z, aq=None, layer_ltype=None):
        if aq is None: aq = self.aq.find_aquifer_data(x, y)
        assert z <= aq.z[0] and z >= aq.z[-1], "z value not inside aquifer"
        if layer_ltype is None:
            layer, ltype, dummy = aq.findlayer(z)
        else:
            layer, ltype = layer_ltype
        h = self.head(x, y, aq=aq)
        # qz between aquifer layers
        qzlayer = np.zeros(aq.naq + 1)
        qzlayer[1:-1] = (h[1:] - h[:-1]) / aq.c[1:]
        if aq.ltype[0] == 'l':
            qzlayer[0] = (h[0] - aq.hstar) / aq.c[0]
        if ltype == 'l':
            vz = qzlayer[layer] / aq.nporll[layer]
            vx = 0
            vy = 0
        else:
            qzbot = qzlayer[layer + 1]
            qztop = qzlayer[layer]
            if layer == 0:
                qztop += self.qztop(x, y)   
            vz = (qzbot + (z - aq.zaqbot[layer]) / aq.Haq[layer] * \
                 (qztop - qzbot)) / aq.nporaq[layer]        
            qx, qy = self.disvec(x, y, aq=aq)
            vx = qx[layer] / (aq.Haq[layer] * aq.nporaq[layer])
            vy = qy[layer] / (aq.Haq[layer] * aq.nporaq[layer])
        return np.array([vx, vy, vz])
               
    def solve(self, printmat=0, sendback=0, silent=False):
        '''Compute solution'''
        # Initialize elements
        self.initialize()
        # Compute number of equations
        self.neq = np.sum([e.nunknowns for e in self.elementlist])
        if self.neq == 0: return
        if silent is False:
            print('Number of elements, Number of equations:', len(
                self.elementlist), ',', self.neq)
        if self.neq == 0:
            if silent is False: print('No unknowns. Solution complete')
            return
        mat = np.empty((self.neq, self.neq))
        rhs = np.empty(self.neq)
        ieq = 0
        for e in self.elementlist:
            if e.nunknowns > 0:
                mat[ieq:ieq + e.nunknowns, :], rhs[ieq:ieq + e.nunknowns] = \
                e.equation()
                ieq += e.nunknowns
            if silent is False:
                print('.', end='', flush=True)
        if printmat:
            return mat, rhs
        sol = np.linalg.solve(mat, rhs)
        icount = 0
        for e in self.elementlist:
            if e.nunknowns > 0:
                e.setparams(sol[icount:icount + e.nunknowns])
                icount += e.nunknowns
        if silent is False:
            print()  # needed cause the dots are printed
            print('solution complete')
        elif (silent == 'dot') or (silent == '.'):
            print('.', end='', flush=True)
        if sendback:
            return sol
        return

    def solve_mp(self, nproc=4, printmat=0, sendback=0, silent=False):
        '''Compute solution, multiprocessing implementation.
        Note: estimated speedup approximately by factor of
        number of physical cores. Virtual cores do not improve
        calculation time.'''
        # Initialize elements
        self.initialize()
        # Compute number of equations
        self.neq = np.sum([e.nunknowns for e in self.elementlist])
        if self.neq == 0: return
        if silent is False:
            print('Number of elements, Number of equations:', len(
                self.elementlist), ',', self.neq)
        if self.neq == 0:
            if silent is False: print('No unknowns. Solution complete')
            return
        mat = np.empty((self.neq, self.neq))
        rhs = np.empty(self.neq)

        # start multiprocessing
        if nproc is None:
            nproc = mp.cpu_count() - 1  # make no. of processes equal to 1 less than no. of cores
        elif nproc > mp.cpu_count():
            print("Given 'nproc' larger than no. of cores on machine. Setting 'nproc' to {}.".format(mp.cpu_count()))
            nproc = mp.cpu_count()

        pool = mp.Pool(processes=nproc)
        results = []
        for e in self.elementlist:
            if e.nunknowns > 0:
                results.append(pool.apply_async(e.equation))
            if silent is False:
                print('.', end='', flush=True)

        pool.close()
        pool.join()

        mat = np.empty((self.neq, self.neq))
        rhs = np.zeros(self.neq)

        ieq = 0

        for p in results:
            imat, irhs = p.get()
            mat[ieq:ieq + imat.shape[0], :] = imat
            rhs[ieq:ieq + irhs.shape[0]] = irhs
            ieq += imat.shape[0]

        # end multiprocessing

        if printmat:
            return mat, rhs
        sol = np.linalg.solve(mat, rhs)
        icount = 0
        for e in self.elementlist:
            if e.nunknowns > 0:
                e.setparams(sol[icount:icount + e.nunknowns])
                icount += e.nunknowns
        if silent is False:
            print()  # needed cause the dots are printed
            print('solution complete')
        elif (silent == 'dot') or (silent == '.'):
            print('.', end='', flush=True)
        if sendback:
            return sol
        return
    
    def write(self):
        rv = self.modelname + ' = ' + self.name + '(\n'
        for key in self.inputargs[1:]:  # The first argument (self) is ignored
            if isinstance(self.inputvalues[key], np.ndarray):
                rv += key + ' = ' + np.array2string(self.inputvalues[key], 
                                                    separator=',') + ',\n'
            elif isinstance(self.inputvalues[key],str):                
                rv += key + " = '" + self.inputvalues[key] + "',\n"
            else:
                rv += key + ' = ' + str(self.inputvalues[key]) + ',\n'
        rv += ')\n'
        return rv
    
    def writemodel(self, fname):
        self.initialize()  # So that the model can be written without solving first
        f = open(fname, 'w')
        f.write('from timml import *\n')
        f.write(self.write())
        for e in self.elementlist:
            f.write(e.write())
        f.close()
        
class ModelMaq(Model):
    """
    Create a Model object by specifying a mult-aquifer sequence of
    aquifer-leakylayer-aquifer-leakylayer-aquifer etc
    
    Parameters
    ----------
    kaq : float, array or list
        Hydraulic conductivity of each aquifer from the top down.
        If float, hydraulic conductivity is the same in all aquifers.
    z : array or list
        Elevation of tops and bottoms of the aquifers from the top down.
        Leaky layers may have zero thickness.
           * if topboundary='conf': length is 2 * number of aquifers
           * if topboundary='semi': length is 2 * number of aquifers + 1 
             as top of leaky layer on top of systems needs to be specified
    c : float, array or list
        Resistance of leaky layers from the top down.
           * if float, resistance is the same for all leaky layers
           * if topboundary='conf': length is number of aquifers - 1
           * if topboundary='semi': length is number of aquifers
    npor : float, array or list
        Porosity of all aquifers and leaky layers from the top down.
           * if float, porosity is the same for all layers
           * if topboundary='conf': length is 2 * number of aquifers - 1
           * if topboundary='semi': length is 2 * number of aquifers
    topboundary : string, 'conf' or 'semi' (default is 'conf')
        Indicates whether the topboundary is confined ('conf') or
        semi-confined ('semi').
    hstar : float or None (default is None)
        Head value above semi-confining top, only read if topboundary='semi'.

    Examples
    --------
    >>> ml = ModelMaq(kaq=[10, 20], z=[20, 12, 10, 0], c=1000)
    
    """
    
    def __init__(self, kaq=1, z=[1, 0], c=[], npor=0.3, topboundary='conf',
                 hstar=None, f2py=False):
        self.storeinput(inspect.currentframe())
        kaq, c, npor, ltype = param_maq(kaq, z, c, npor, topboundary)
        Model.__init__(self, kaq, c, z, npor, ltype, f2py)
        self.name = 'ModelMaq'
        if self.aq.ltype[0] == 'l':
            ConstantStar(self, hstar, aq=self.aq)
            
class Model3D(Model):
    """
    Model3D Class to create a multi-layer model object consisting of
    many aquifer layers. The resistance between the layers is computed
    from the vertical hydraulic conductivity of the layers.
    
    Parameters
    ----------
    kaq : float, array or list
        hydraulic conductivity of each layer from the top down
        if float, hydraulic conductivity is the same in all aquifers
    z : array or list
        elevation of top of system followed by bottoms of all layers
        from the top down
        bottom of layer is automatically equal to top of layer below it
        length is number of aquifer layers + 1
    kzoverkh : float
        vertical anisotropy ratio vertical k divided by horizontal k
        if float, value is the same for all layers
        length is number of layers
    npor : float, array or list
        porosity of all aquifer layers
        from the top down
        if float, porosity is the same for all layers
        if topboundary='conf': length is number of layers
        if topboundary='semi': length is number of layers + 1
    topboundary : string, 'conf' or 'semi' (default is 'conf')
        indicating whether the top is confined ('conf') or
        semi-confined ('semi')
    topres : float
        resistance of top semi-confining layer (read if topboundary='semi')
    topthick: float
        thickness of top semi-confining layer (read if topboundary='semi')
    hstar : float or None (default is None)
        head value above semi-confining top (read if topboundary='semi')

    Examples
    --------
    >>> ml = Model3D(kaq=10, z=np.arange(20, -1, -2), kzoverkh=0.1)
    
    """
    
    def __init__(self, kaq=1, z=[1, 0], kzoverkh=1, npor=0.3,
                 topboundary='conf', topres=0, topthick=0, hstar=0,
                 f2py=False):
        '''Model3D
        for semi-confined aquifers, set top equal to 'semi' and provide
        topres: resistance of top
        tophick: thickness of top
        hstar: head above top'''
        self.storeinput(inspect.currentframe())
        kaq, c, npor, ltype = param_3d(kaq, z, kzoverkh, npor, topboundary,
                                       topres)
        if topboundary == 'semi':
            z = np.hstack((z[0] + topthick, z))
        Model.__init__(self, kaq, c, z, npor, ltype, f2py)
        self.name = 'Model3D'
        if self.aq.ltype[0] == 'l':
            ConstantStar(self, hstar, aq=self.aq)




'''
Copyright (C), 2015, Mark Bakker.
Mark Bakker, Delft University of Technology
mark dot bakker at tudelft dot nl

TimML is a computer program for the simulation of steady-state
multiaquifer flow with analytic elements and consists of a
library of Python scripts and FORTRAN extensions.
'''
#from __future__ import division, print_function, absolute_import

#--version number
__name__='timml'
__author__='Mark Bakker'
from .version import __version__

# Import all classes and functions
from .model import ModelMaq, Model3D, Model
from .well import WellBase, Well, HeadWell
from .constant import Constant, ConstantStar
from .linesink import LineSinkBase, HeadLineSinkZero, HeadLineSink, \
                      LineSinkDitch, HeadLineSinkString, LineSinkDitchString, \
                      HeadLineSinkContainer
from .linedoublet import ImpLineDoublet, ImpLineDoubletString, \
                         LeakyLineDoublet, LeakyLineDoubletString
from .circareasink import CircAreaSink
from .inhomogeneity import PolygonInhomMaq, BuildingPit
from .inhomogeneity1d import StripInhomMaq, StripInhom3D
from .uflow import Uflow
from .trace import timtraceline, timtracelines
from .linesink1d import LineSink1D, HeadLineSink1D
from .linedoublet1d import ImpLineDoublet1D, LeakyLineDoublet1D
from .stripareasink import StripAreaSink

__all__ = [s for s in dir() if not s.startswith("_")]



'''
mlcircinhom.py contains the CircleInhom class
This file is part of the TimML library and is distributed under
the GNU LPGL. See the TimML.py file for more details.
(c) Mark Bakker, 2002-2007
'''

from numpy import *
import scipy.special
from mlelement import *

class CircleInhom(Element):
    '''CircleInhom class
    Note that aquiferparent doesn't have a meaning for this element
    All attributes from element.
    '''
    Rconvsq = 7 * 7
    def __init__(self,modelParent,order,aqin,aqout,label=None):
	Element.__init__(self,modelParent)
	self.order = order
        self.aqin = aqin; self.aqout = aqout
        assert aqin.Naquifers == aqout.Naquifers, "TimML Input error: Number of aquifers inside and outside must be equal"
        self.label = label
        self.type = 'circleinhom'
	self.setCoefs()
        self.modelParent.addElement(self)
    def __repr__(self):
	return 'CircInhom xc,yc,order ' + str((self.xc,self.yc,self.order)) + ' with inside aquifer data ' + str(self.aqin)
    def setCoefs(self):
        self.xc = self.aqin.xc; self.yc = self.aqin.yc; self.R = self.aqin.R; self.Rsq = self.aqin.Rsq
        self.paramin = zeros((2*self.order+1,self.aqin.Naquifers),'d') 
        self.paramout = zeros((2*self.order+1,self.aqin.Naquifers),'d')
        self.Nparamin = ( 2*self.order+1 ) * self.aqin.Naquifers
        self.Nparamout = self.Nparamin
        # Compute control points
        self.Ncp = 2 * self.order + 1
        self.thetacp = arange( 0, 2*pi-0.01, 2*pi/self.Ncp )
        self.xcp = self.xc + self.R * cos(self.thetacp)
        self.ycp = self.yc + self.R * sin(self.thetacp)
        # Compute values on the edge (to scale functions)
        self.matRin = zeros( (self.order+1,self.aqin.Naquifers), 'd' )  # values on inside edge
        rolab = self.R/self.aqin.lab  # vector with all R/labin values
        if self.aqin.type == self.aqin.conf:
            for p in range(self.order+1):
                self.matRin[p,0] = self.R**p  # first value in row is R**p
                self.matRin[p,1:] = scipy.special.iv(p,rolab) # other values are Ip(R/lab)
        elif self.aqin.type == self.aqin.semi:
             for p in range(self.order+1):
                self.matRin[p,:] = scipy.special.iv(p,rolab) # all values are Ip(R/lab)
        self.matRin = 1.0 / self.matRin
        self.matRout = zeros( (self.order+1,self.aqout.Naquifers), 'd' )  # values on outside edge
        rolab = self.R/self.aqout.lab  # vector with all R/labout values
        if self.aqout.type == self.aqin.conf:
            for p in range(self.order+1):
                self.matRout[p,0] = 1./self.R**(p)  # first value in row is 1/r**p
                self.matRout[p,1:] = scipy.special.kn(p,rolab) # other values are Kp
            self.matRout[0,0] = log(self.R+1.0) / (2.0*pi)  # first term is logarithm, so scale by log(self.R+1)to make sure that self.R can be 1
        elif self.aqout.type == self.aqout.semi:
            for p in range(self.order+1):
                self.matRout[p,:] = scipy.special.kn(p,rolab) # other values are Kp
        self.matRout = 1.0 / self.matRout
    def potentialInfluence(self,aq,x,y):
        r = sqrt( (x-self.xc)**2 + (y-self.yc)**2 )
        theta = arctan2( (y-self.yc),(x-self.xc) )
        mat = zeros( (self.order+1,aq.Naquifers), 'd' )  # matrix to store values
     # Inside Cylinder
        if aq == self.aqin:
            rolab = r/aq.lab  # vector with all r/lab values
            if aq.type == aq.conf:  # Confined aquifer
                for p in range(self.order+1):
                    mat[p,0] = r**p  # first value in row is r**p
                    mat[p,1:] = scipy.special.iv(p,rolab) # other values are Ip
            elif aq.type == aq.semi:  # Semi-confined aquifer
                for p in range(self.order+1):
                    mat[p,:] = scipy.special.iv(p,rolab) # all values are Ip
            mat = mat * self.matRin
     # In aquifer outside cylinder
        elif aq == self.aqout:
            rolab = r/aq.lab  # vector with all r/lab values
            if aq.type == aq.conf:  # Confined aquifer
                for p in range(self.order+1):
                    mat[p,0] = 1./r**(p)  # first value in row is 1/r**p
                    mat[p,1:] = scipy.special.kn(p,rolab) # other values are Kp
                mat[0,0] = log(r) / (2.0*pi)  # logarithm on outside
            if aq.type == aq.semi:  # Semi-confined aquifer
                for p in range(self.order+1):
                    mat[p,:] = scipy.special.kn(p,rolab) # all values are Kp                
            mat = mat * self.matRout
     # Somewhere else, assume outside
        else:
            if self.aqout.type == self.aqout.conf and aq.type == aq.conf:
                for p in range(self.order+1):
                    mat[p,0] = 1./r**(p)  # first value in row is 1/r**p
                mat[0,0] = log(r) / (2.0*pi)
                mat = mat * self.matRout
     # Store values in return value
        rv = zeros( (2*self.order+1,aq.Naquifers), 'd' )
        rv[0,:] = mat[0,:]  # First row only has cosine part; sin(0)=0
        for p in range(1,self.order+1):
            rv[2*p-1,:] = mat[p,:] * cos( p*theta )
            rv[2*p,:]   = mat[p,:] * sin( p*theta )
        return rv
    def potentialInfluenceInLayer(self,aq,pylayer,x,y):
        '''Returns PotentialInfluence function in aquifer aq in pylayer as array (1 value per parameter)
        Needs to be overloaded because there is no parameter outside the functions
        Needs to be modified for inside and outside'''
        potInf = self.potentialInfluence(aq,x,y)
        rv = zeros(0,'d')
        for p in potInf:
            rv = hstack(( rv, p * aq.eigvec[pylayer,:] ))
        if aq == self.aqin:
            rv = hstack(( rv, zeros( self.Nparamout, 'd' ) ))
        else:
            rv = hstack(( zeros( self.Nparamin, 'd' ), rv ))
        return rv
    def potentialInfluenceAllLayers(self,aq,pylayer,x,y):
        '''Returns PotentialInfluence function in aquifer aq in all layers as an array'''
        potInf = self.potentialInfluence(aq,x,y)
        rv = zeros((aq.Naquifers,0),'d')
        for p in potInf:
            rv = hstack( ( rv, p*aq.eigvec ) )
        if aq == self.aqin:
            rv = hstack(( rv, zeros( ( aq.Naquifers, self.Nparamout ), 'd' ) ))
        else:
            rv = hstack(( zeros( (aq.Naquifers,self.Nparamin), 'd' ), rv ))
        return rv
    def potentialInfluenceSpecLayers(self,aq,pylayer,x,y):
        '''Returns PotentialInfluence function in aquifer aq in all layers as an array'''
        potInf = self.potentialInfluence(aq,x,y)
        pylen = len(pylayer)
        rv = zeros((pylen,0),'d')
        eigvec = take( aq.eigvec, pylayer, 0 )
        for p in potInf:
            rv = hstack( ( rv, p*eigvec ) )
        if aq == self.aqin:
            rv = hstack(( rv, zeros( ( pylen, self.Nparamout ), 'd' ) ))
        else:
            rv = hstack(( zeros( (pylen,self.Nparamin), 'd' ), rv ))
        return rv
    def potentialContribution(self,aq,x,y):
        '''Returns array of potentialContribution. Needs to be overloaded cause there is inside and outside'''
        if aq == self.aqin:
            potInf = sum( self.paramin * self.potentialInfluence(aq,x,y), 0 )
        else:
            potInf = sum( self.paramout * self.potentialInfluence(aq,x,y), 0 )
        return potInf
    def dischargeInfluence(self,aq,x,y):
        r = sqrt( (x-self.xc)**2 + (y-self.yc)**2 )
        theta = arctan2( (y-self.yc),(x-self.xc) )
        mat = zeros( (self.order+1,aq.Naquifers), 'd' )  # matrix to store values
        if aq == self.aqin:  # Inside cylinder
            rolab = r/aq.lab  # vector with all r/lab values
            if aq.type == aq.conf:  # Confined aquifer
                for p in range(self.order+1):
                    if p == 0:
                        mat[0,0] = 0.0
                    else:
                        mat[p,0] = - float(p) * r**(p-1)  # first value in row is pr**(p-1)
                    mat[p,1:] = - ( scipy.special.iv(p-1,rolab) + scipy.special.iv(p+1,rolab) ) / (2 * aq.lab)
            elif aq.type == aq.semi:
                for p in range(self.order+1):
                    mat[p,:] = - ( scipy.special.iv(p-1,rolab) + scipy.special.iv(p+1,rolab) ) / (2 * aq.lab)
            mat = mat * self.matRin
        elif aq == self.aqout:  # In aquifer outside cylinder
            rolab = r/aq.lab  # vector with all r/lab values
            if aq.type == aq.conf:
                for p in range(self.order+1):
                    mat[p,0] = p/r**(p+1)  # first value in row is 1/r**p
                    mat[p,1:] = ( scipy.special.kn(p-1,rolab) + scipy.special.kn(p+1,rolab) ) / (2 * aq.lab) 
                mat[0,0] = - 1.0 / (2.0*pi*r)  # log
            elif aq.type == aq.semi:
                for p in range(self.order+1):
                    mat[p,:] = ( scipy.special.kn(p-1,rolab) + scipy.special.kn(p+1,rolab) ) / (2 * aq.lab) 
            mat = mat * self.matRout
        else:  # Somewhere else
            if self.aqout.type == self.aqout.conf and aq.type == aq.conf:
                mat[0,0] = - 1.0 / (2.0*pi*r)
                for p in range(1,self.order+1):
                    mat[p,0] = p/r**(p+1)  # first value in row is 1/r**p
                mat = mat * self.matRout
        # Store values in return value
        rvrad = zeros( (2*self.order+1,aq.Naquifers), 'd' )
        rvrad[0,:] = mat[0,:]  # First row only has cosine part; sin(0)=0
        for p in range(1,self.order+1):
            rvrad[2*p-1,:] = mat[p,:] * cos( p*theta )
            rvrad[2*p,:]   = mat[p,:] * sin( p*theta )
    # Tangential part
        mat = zeros( (self.order+1,aq.Naquifers), 'd' )  # matrix to store values
        if aq == self.aqin:  # Inside cylinder
            rolab = r/aq.lab  # vector with all r/lab values
            if aq.type == aq.conf:
               for p in range(1,self.order+1):  # all values for p=0 are zero (no dependence on theta)
                    mat[p,0] = float(p) * r**(p-1)  # first value in row is pr**(p-1)
                    mat[p,1:] = float(p) * scipy.special.iv(p,rolab) / r
            elif aq.type == aq.semi:
               for p in range(1,self.order+1):  # all values for p=0 are zero (no dependence on theta)
                    mat[p,:] = float(p) * scipy.special.iv(p,rolab) / r
            mat = mat * self.matRin
        elif aq == self.aqout:  # In aquifer outside cylinder
            rolab = r/aq.lab  # vector with all r/lab values
            if aq.type == aq.conf:
                for p in range(1,self.order+1):  # all values for p=0 are zero (no dependence on theta)
                    mat[p,0] = float(p)/r**(p+1)  # first value in row is 1/r**p
                    mat[p,1:] = float(p) * scipy.special.kn(p,rolab) / r
            elif aq.type == aq.semi:
                for p in range(1,self.order+1):  # all values for p=0 are zero (no dependence on theta)
                    mat[p,:] = float(p) * scipy.special.kn(p,rolab) / r
            mat = mat * self.matRout
        else:  # Somewhere else
            if self.aqout.type == self.aqout.conf and aq.type == aq.conf:
                for p in range(1,self.order+1):
                    mat[p,0] = float(p)/r**(p+1)
            mat = mat * self.matRout
        # Store values in return value
        rvtan = zeros( (2*self.order+1,aq.Naquifers), 'd' )
        for p in range(1,self.order+1):
            rvtan[2*p-1,:] = mat[p,:] * sin( p*theta )
            rvtan[2*p,:]   = - mat[p,:] * cos( p*theta )
        qx = rvrad * cos(theta) - rvtan * sin(theta);
        qy = rvrad * sin(theta) + rvtan * cos(theta);
        return [qx,qy]
    def dischargeInfluenceRad(self,aq,x,y):
        r = sqrt( (x-self.xc)**2 + (y-self.yc)**2 )
        theta = arctan2( (y-self.yc),(x-self.xc) )
        mat = zeros( (self.order+1,aq.Naquifers), 'd' )  # matrix to store values
        if aq == self.aqin:  # Inside cylinder
            rolab = r/aq.lab  # vector with all r/lab values
            for p in range(self.order+1):
                if p == 0:
                    mat[0,0] = 0.0
                else:
                    mat[p,0] = - float(p) * r**(p-1)  # first value in row is pr**(p-1)
                mat[p,1:] = - ( scipy.special.iv(p-1,rolab) + scipy.special.iv(p+1,rolab) ) / (2 * aq.lab)
            mat = mat * self.matRin
        elif aq == self.aqout:  # In aquifer outside cylinder
            rolab = r/aq.lab  # vector with all r/lab values
            for p in range(self.order+1):
                mat[p,0] = p/r**(p+1)  # first value in row is 1/r**p
                mat[p,1:] = ( scipy.special.kn(p-1,rolab) + scipy.special.kn(p+1,rolab) ) / (2 * aq.lab) 
            mat[0,0] = 0.0  # no constant on outside
            mat = mat * self.matRout
        else:  # Somewhere else
            for p in range(self.order+1):
                mat[p,0] = p/r**(p+1)  # first value in row is 1/r**p
            mat[0,0] = 0.0
            mat = mat * self.matRout
        # Store values in return value
        rvrad = zeros( (2*self.order+1,aq.Naquifers), 'd' )
        rvrad[0,:] = mat[0,:]  # First row only has cosine part; sin(0)=0
        for p in range(1,self.order+1):
            rvrad[2*p-1,:] = mat[p,:] * cos( p*theta )
            rvrad[2*p,:]   = mat[p,:] * sin( p*theta )
        return rvrad
    def dischargeInfluenceRadInLayer(self,aq,pylayer,x,y):
        '''Returns dischargeInfluenceRadInLayer function in aquifer aq in pylayer as list (1 value per parameter)
        Needs to be overloaded because there is no parameter outside the functions
        Needs to be modified for inside and outside'''
        disInf = self.dischargeInfluenceRad(aq,x,y)
        rv = []
        for d in disInf:
            rv = rv + list(d * aq.eigvec[pylayer,:])
        if aq != self.aqin:
            dump = rv.pop(0)  # Get rid of very first term (always zero)
        return rv
    def dischargeInfluenceRadAllLayers(self,aq,dumlayer,x,y):
        '''Returns dischargeInfluenceRadAllLayers function in aquifer aq as an array
        Needs to be overloaded because there is no parameter outside the functions
        Needs to be modified for inside and outside'''
        disInf = self.dischargeInfluenceRad(aq,x,y)
        rv = zeros((aq.Naquifers,0),'d')
        for d in disInf:
            rv = hstack( ( rv, d*aq.eigvec ) )
        if aq != self.aqin:
            rv = rv[:,1:]  # Take off first column (parameter always zero)
        return rv
    def dischargeInfluenceAllLayers(self,aq,dumlayer,x,y):
        '''Returns dischargeInfluenceAllLayers function in aquifer aq as an array
        Needs to be overloaded because there is no parameter outside the functions
        Needs to be modified for inside and outside'''
        [disx,disy] = self.dischargeInfluence(aq,x,y)
        rvx = zeros((aq.Naquifers,0),'d')
        rvy = zeros((aq.Naquifers,0),'d')
        for d in disx:
            rvx = hstack( ( rvx, d*aq.eigvec ) )
        for d in disy:
            rvy = hstack( ( rvy, d*aq.eigvec ) )
        if aq == self.aqin:
            rvx = hstack(( rvx, zeros( ( aq.Naquifers, self.Nparamout ), 'd' ) ))
            rvy = hstack(( rvy, zeros( ( aq.Naquifers, self.Nparamout ), 'd' ) ))
        else:
            rvx = hstack(( zeros( (aq.Naquifers,self.Nparamin), 'd' ), rvx ))
            rvy = hstack(( zeros( (aq.Naquifers,self.Nparamin), 'd' ), rvy ))
        return [rvx,rvy]
    def dischargeInfluenceInLayer(self,aq,pylayer,x,y):
        '''Returns dischargeInfluence in pylayer, modified for paramin and paramout'''
        [disx,disy] = self.dischargeInfluence(aq,x,y)
        rvx = disx * aq.eigvec[pylayer,:]
        rvy = disy * aq.eigvec[pylayer,:]
        if aq == self.aqin:
            rvx = hstack(( rvx.ravel(), zeros( self.Nparamout, 'd' ) ))
            rvy = hstack(( rvy.ravel(), zeros( self.Nparamout, 'd' ) ))
        else:
            rvx = hstack(( zeros( self.Nparamin, 'd' ), rvx.ravel() ))
            rvy = hstack(( zeros( self.Nparamin, 'd' ), rvy.ravel() ))
        return [rvx,rvy]
    def dischargeInfluenceSpecLayers(self,aq,pylayer,x,y):
        '''Returns dischargeInfluenceAllLayers function in aquifer aq as an array
        Needs to be overloaded because there is no parameter outside the functions
        Needs to be modified for inside and outside'''
        [disx,disy] = self.dischargeInfluence(aq,x,y)
        pylen = len(pylayer)
        rvx = zeros((pylen,0),'d')
        rvy = zeros((pylen,0),'d')
        eigvec = take( aq.eigvec, pylayer, 0 )
        for d in disx:
            rvx = hstack( ( rvx, d*eigvec ) )
        for d in disy:
            rvy = hstack( ( rvy, d*eigvec ) )
        if aq == self.aqin:
            rvx = hstack(( rvx, zeros( ( pylen, self.Nparamout ), 'd' ) ))
            rvy = hstack(( rvy, zeros( ( pylen, self.Nparamout ), 'd' ) ))
        else:
            rvx = hstack(( zeros( (pylen,self.Nparamin), 'd' ), rvx ))
            rvy = hstack(( zeros( (pylen,self.Nparamin), 'd' ), rvy ))
        return [rvx,rvy]

    def dischargeContribution(self,aq,x,y):
        '''Returns matrix with two rowvectors of dischargeContributions Qx and Qy. Needs to be overloaded cause there is inside and outside'''
        disInf = self.dischargeInfluence(aq,x,y)
        if aq == self.aqin:
            disxInf = sum( self.paramin * disInf[0], 0 )
            disyInf = sum( self.paramin * disInf[1], 0 )
        else:
            disxInf = sum( self.paramout * disInf[0], 0 )
            disyInf = sum( self.paramout * disInf[1], 0 )
        return array([disxInf,disyInf])
    def zeroFunction(self,aqdum,ldum,xdum,ydum):
        '''Returns list of zeros of length number of parameters'''
        return list( zeros( self.Nparamin+self.Nparamout,'d' ) )
    def getMatrixRows(self,elementList):
        rows=[]
        # Jump in potential
        for i in range(self.Ncp):  # Hardcoded for same number of aquifers in and out
            rowin = zeros((self.aqin.Naquifers,0),'d'); rowout = zeros((self.aqin.Naquifers,0),'d')  # Zero columns!
            for e in elementList:
                rowpart = e.getMatrixCoefficients(self.aqin,self.ldum,self.xcp[i],self.ycp[i],\
                                lambda el,aq,ldum,x,y:el.potentialInfluenceAllLayers(aq,ldum,x,y))
                if size(rowpart) > 0: rowin = hstack(( rowin, rowpart ))
            for e in elementList:
                rowpart = e.getMatrixCoefficients(self.aqout,self.ldum,self.xcp[i],self.ycp[i],\
                                lambda el,aq,ldum,x,y:el.potentialInfluenceAllLayers(aq,ldum,x,y))
                if size(rowpart) > 0: rowout = hstack(( rowout, rowpart ))
            row = self.aqout.Tcol * rowin - self.aqin.Tcol * rowout
            row = hstack( ( row,
                self.aqin.Tcol * self.aqout.Tcol * ( self.aqout.hstar - self.aqin.hstar ) + \
                self.aqin.Tcol * self.modelParent.potentialVectorCol(self.xcp[i],self.ycp[i],self.aqout) - \
                self.aqout.Tcol * self.modelParent.potentialVectorCol(self.xcp[i],self.ycp[i],self.aqin) ) )
            rows = rows + row.tolist()
        # Jump in radial discharge
        for i in range(self.Ncp):  # Hardcoded for same number of aquifers in and out
            rowin = zeros((self.aqin.Naquifers,0),'d'); rowout = zeros((self.aqin.Naquifers,0),'d')  # Zero columns!
            for e in elementList:
                rowqxqy = e.getMatrixCoefficients(self.aqin,self.ldum,self.xcp[i],self.ycp[i],\
                                lambda el,aq,ldum,x,y:el.dischargeInfluenceAllLayers(aq,ldum,x,y))
                if size(rowqxqy) > 0:
                    rowpart = rowqxqy[0] * cos( self.thetacp[i] ) + rowqxqy[1] * sin( self.thetacp[i] )
                    rowin = hstack(( rowin, rowpart ))
            for e in elementList:
                rowqxqy = e.getMatrixCoefficients(self.aqout,self.ldum,self.xcp[i],self.ycp[i],\
                                lambda el,aq,ldum,x,y:el.dischargeInfluenceAllLayers(aq,ldum,x,y))
                if size(rowqxqy) > 0:
                    rowpart = rowqxqy[0] * cos( self.thetacp[i] ) + rowqxqy[1] * sin( self.thetacp[i] )
                    rowout = hstack(( rowout, rowpart ))
            row = rowin - rowout
            row = hstack(( row,
                self.modelParent.dischargeNormVectorCol(self.xcp[i],self.ycp[i],self.thetacp[i],self.aqout) -\
                self.modelParent.dischargeNormVectorCol(self.xcp[i],self.ycp[i],self.thetacp[i],self.aqin) ))
            rows = rows + row.tolist()
        return rows
    def getMatrixCoefficients(self,aq,pylayer,x,y,func):
        return func(self,aq,pylayer,x,y)
    def takeParameters(self,xsol,icount):
        par = xsol[icount : icount + self.Nparamin]
        self.paramin = self.paramin + reshape(par,(2*self.order+1,self.aqin.Naquifers))
        icount = icount + self.Nparamin
        par = xsol[icount : icount + self.Nparamout]
        self.paramout = self.paramout + reshape(par,(2*self.order+1,self.aqin.Naquifers))
        icount = icount + self.Nparamout
        return icount
    def check(self):
        for i in range(self.Ncp):
            print 'Control point '+str(i)
            print 'head inside:  '+str(self.modelParent.headVector(self.xcp[i],self.ycp[i],self.aqin))
            print 'head outside: '+str(self.modelParent.headVector(self.xcp[i],self.ycp[i],self.aqout))
            print 'Qrad inside:  '+str(self.modelParent.dischargeNormVector(self.xcp[i],self.ycp[i],self.thetacp[i],self.aqin))
            print 'Qrad outside: '+str(self.modelParent.dischargeNormVector(self.xcp[i],self.ycp[i],self.thetacp[i],self.aqout))
    def layout(self):
        return [0]
    def parSum(self):
        return sum( sum(abs(self.paramin)) ) + sum( sum(abs(self.paramout)) )
    def nearElement(self,pyLayer,xyz1,xyz2,step,idir):
        changed = 0; stop = 0; xyznew = zeros(3,'d')
        r1sq = ( xyz1[0] - self.xc )**2 + ( xyz1[1] - self.yc ) **2
        r2sq = ( xyz2[0] - self.xc )**2 + ( xyz2[1] - self.yc ) **2
        if ( r1sq < self.Rsq and r2sq > self.Rsq  ) or ( r1sq > self.Rsq and r2sq < self.Rsq  ):
            (x1,y1) = xyz1[0:2]; (x2,y2) = xyz2[0:2]
            a = (x2-x1)**2 + (y2-y1)**2
            b = 2.0 * ( (x2-x1) * (x1-self.xc) + (y2-y1) * (y1-self.yc) )
            c = self.xc**2 + self.yc**2 + x1**2 + y1**2 - 2.0 * (self.xc*x1 +self.yc*y1) - self.Rsq
            u1 = ( -b - sqrt(b**2 - 4.0*a*c) ) / (2.0*a)
            u2 = ( -b + sqrt(b**2 - 4.0*a*c) ) / (2.0*a)
            if u1 > 0:
                u = u1 * 1.000001 # Go just beyond circle
            else:
                u = u2 * 1.000001 # Go just beyond circle
            xn = x1 + u * (x2-x1); yn = y1 + u * (y2-y1)
            zn = xyz1[2] + u * ( xyz2[2] - xyz1[2] )
            xyznew = array([xn,yn,zn],'d')
            changed = 1
        return [changed, stop, xyznew] 
    def distanceSquaredToElement(self,x,y):
        '''Returns distance squared to element. Used for deciding tracing step'''
        dis = sqrt( ( x - self.xc )**2 + ( y - self.yc ) **2 )
        dissq = ( dis - self.R ) **2
        return dissq



import numpy as np
import matplotlib.pyplot as plt
import inspect  # Used for storing the input
from .element import Element
from .equation import HeadEquation, PotentialEquation
from .besselaesnumba import besselaesnumba
besselaesnumba.initialize()
try:
    from .src import besselaesnew
    besselaesnew.besselaesnew.initialize()
    #print('succes on f2py')
except:
    pass

from .controlpoints import controlpoints, strengthinf_controlpoints

__all__ = ['LineSinkBase', 'HeadLineSinkZero', 'HeadLineSink', 'LineSinkDitch',
           'HeadLineSinkString', 'LineSinkDitchString']

class LineSinkChangeTrace:
    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer, direction, hstepmax, verbose=False):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        if (ltype == 'a'):
            if True:
#            if (layer == self.layers).any():  # in layer where line-sink is screened
# not needed anymore, I thin this is all taken care of with checking Qn1 and Qn2
                if verbose:
                    print('hello changetrace')
                    print('xyz1:', xyzt1[:-1])
                    print('xyz2:', xyzt2[:-1])
                x1, y1, z1, t1 = xyzt1
                x2, y2, z2, t2 = xyzt2
                eps = 1e-8
                za = x1 + y1 * 1j
                zb = x2 + y2 * 1j
                Za = (2 * za - (self.z1 + self.z2)) / (self.z2 - self.z1)
                Zb = (2 * zb - (self.z1 + self.z2)) / (self.z2 - self.z1)
                if verbose:
                    print('Za', Za)
                    print('Zb', Zb)
                if Za.imag * Zb.imag < 0:
                    Xa, Ya = Za.real, Za.imag
                    Xb, Yb = Zb.real, Zb.imag
                    X = Xa - Ya * (Xb - Xa) / (Yb - Ya)
                    if verbose: print('X', X)
                    if abs(X) <= 1:  # crosses line-sink
                        if verbose: print('crosses line-sink')
                        Znew1 = X - eps * np.sign(Yb) * 1j  # steps to side of Ya
                        Znew2 = X + eps * np.sign(Yb) * 1j  # steps to side of Yb
                        znew1 = 0.5 * ((self.z2 - self.z1) * Znew1 + self.z1 + self.z2)
                        znew2 = 0.5 * ((self.z2 - self.z1) * Znew2 + self.z1 + self.z2)
                        xnew1, ynew1 = znew1.real, znew1.imag
                        xnew2, ynew2 = znew2.real, znew2.imag
                        if Ya < 0:
                            theta = self.theta_norm_out
                        else:
                            theta = self.theta_norm_out + np.pi
                        Qx1, Qy1 = self.model.disvec(xnew1, ynew1)[:, layer] * direction
                        Qn1 = Qx1 * np.cos(theta) + Qy1 * np.sin(theta)
                        Qx2, Qy2 = self.model.disvec(xnew2, ynew2)[:, layer] * direction
                        Qn2 = Qx2 * np.cos(theta) + Qy2 * np.sin(theta)
                        if verbose:
                            print('xnew1, ynew1:', xnew1, ynew1)
                            print('xnew2, ynew2:', xnew2, ynew2)
                            print('Qn1, Qn2', Qn1, Qn2)
                            print('Qn2 > Qn1:', Qn2 > Qn1)
                        if Qn1 < 0:  # trying to cross line-sink that infiltrates, stay on bottom, don't terminate
                            if verbose: print('change 1')
                            xnew = xnew1
                            ynew = ynew1
                            dold = np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)
                            dnew = np.sqrt((x1 - xnew) ** 2 + (y1 - ynew) ** 2)
                            znew = z1 + dnew / dold * (z2 - z1)
                            tnew = t1 + dnew / dold * (t2 - t1)
                            changed = True
                            xyztnew = [np.array([xnew, ynew, znew, tnew])]
                        elif Qn2 < 0:  # all water is taken out, terminate
                            if verbose: print('change 2')
                            xnew = xnew2
                            ynew = ynew2
                            dold = np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)
                            dnew = np.sqrt((x1 - xnew) ** 2 + (y1 - ynew) ** 2)
                            znew = z1 + dnew / dold * (z2 - z1)
                            tnew = t1 + dnew / dold * (t2 - t1)
                            changed = True
                            terminate = True
                            xyztnew = [np.array([xnew, ynew, znew, tnew])]
                        elif Qn2 > Qn1:  # line-sink infiltrates
                            if verbose: print('change 3')
                            xnew = xnew2
                            ynew = ynew2
                            dold = np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)
                            dnew = np.sqrt((x1 - xnew) ** 2 + (y1 - ynew) ** 2)
                            znew = z1 + dnew / dold * (z2 - z1)  # elevation just before jump
                            tnew = t1 + dnew / dold * (t2 - t1)
                            Qbelow = (znew - aq.z[modellayer + 1]) / aq.Haq[layer] * Qn1
                            znew2 = aq.z[modellayer + 1] + Qbelow / Qn2 * aq.Haq[layer]
                            changed = True
                            xyztnew = [np.array([xnew, ynew, znew, tnew]), 
                                       np.array([xnew, ynew, znew2, tnew])]
                        else:  # line-sink takes part of water out
                            if verbose: print('change 4')
                            xnew = xnew2
                            ynew = ynew2
                            dold = np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)
                            dnew = np.sqrt((x1 - xnew) ** 2 + (y1 - ynew) ** 2)
                            znew = z1 + dnew / dold * (z2 - z1)  # elevation just before jump
                            tnew = t1 + dnew / dold * (t2 - t1)
                            Qbelow = (znew - aq.z[modellayer + 1]) / aq.Haq[layer] * Qn1
                            if Qbelow > Qn2:  # taken out
                                terminate = True
                                xyztnew = [np.array([xnew, ynew, znew, tnew])]
                            else:
                                znew2 = aq.z[modellayer + 1] + Qbelow / Qn2 * aq.Haq[layer]
                                xyztnew = [np.array([xnew, ynew, znew, tnew]), np.array([xnew, ynew, znew2, tnew])]
                            changed = True
        if terminate:
            message = 'reached element of type linesink'
            if self.label is not None:
                message += ' ({lab})'.format(lab=self.label)
        if verbose:
            print('xyztnew, changed', xyztnew, changed)
        return changed, terminate, xyztnew, message


class LineSinkBase(LineSinkChangeTrace, Element):
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, Qls=100.0, \
                 res=0, wh=1, layers=0, name='LineSinkBase', label=None, \
                 addtomodel=True):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers, \
                         name=name, label=label)
        self.nparam = len(self.layers)
        self.x1 = float(x1)
        self.y1 = float(y1)
        self.x2 = float(x2)
        self.y2 = float(y2)
        self.Qls = np.atleast_1d(Qls)
        self.res = float(res)
        self.wh = wh
        self.addtomodel = addtomodel
        if self.addtomodel: self.model.add_element(self)
        if self.model.f2py:
            self.bessel = besselaesnew.besselaesnew
        else:
            self.bessel = besselaesnumba

    def __repr__(self):
        return self.name + ' from ' + str((self.x1, self.y1)) + ' to ' + str(
            (self.x2, self.y2))

    def initialize(self):
        self.xc = np.array([0.5 * (self.x1 + self.x2)])
        self.yc = np.array([0.5 * (self.y1 + self.y2)])
        self.ncp = 1
        self.z1 = self.x1 + 1j * self.y1
        self.z2 = self.x2 + 1j * self.y2
        self.L = np.abs(self.z1 - self.z2)
        self.theta_norm_out = np.arctan2(self.y2 - self.y1,
                                         self.x2 - self.x1) + np.pi / 2
        self.order = 0  # This is for uniform strength only
        self.aq = self.model.aq.find_aquifer_data(self.xc, self.yc)
        if self.addtomodel: self.aq.add_element(self)
        self.parameters = np.empty((self.nparam, 1))
        self.parameters[:, 0] = self.Qls / self.L
        if self.wh == 'H':
            self.whfac = self.aq.Haq[self.layers]
        elif self.wh == '2H':
            self.whfac = 2.0 * self.aq.Haq[self.layers]
        elif np.isscalar(self.wh):
            self.whfac = self.wh * np.ones(self.nlayers)
        self.resfac = self.aq.T[self.layers] * self.res / self.whfac

    def potinf(self, x, y, aq=None):
        '''Can be called with only one x,y value'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq))
        if aq == self.aq:
            pot = np.zeros(aq.naq)
            pot[:] = self.bessel.potbeslsho(float(x), float(y), self.z1,
                self.z2, aq.lab, 0, aq.ilap, aq.naq)
            rv[:] = self.aq.coef[self.layers] * pot
        return rv

    def disvecinf(self, x, y, aq=None):
        '''Can be called with only one x,y value'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nparam, aq.naq))
        if aq == self.aq:
            qxqy = np.zeros((2, aq.naq))
            qxqy[:, :] = self.bessel.disbeslsho(float(x), float(y), self.z1,
                self.z2, aq.lab, 0, aq.ilap, aq.naq)
            rv[0] = self.aq.coef[self.layers] * qxqy[0]
            rv[1] = self.aq.coef[self.layers] * qxqy[1]
        return rv

    def discharge(self):
        # returns the discharge in each layer
        Q = np.zeros(self.aq.naq)
        Q[self.layers] = self.parameters[:, 0] * self.L
        return Q

    def plot(self, layer=None):
        if (layer is None) or (layer in self.layers):
            plt.plot([self.x1, self.x2], [self.y1, self.y2], 'k')

class HeadLineSinkZero(LineSinkBase, HeadEquation):
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, hls=1.0, \
                 res=0, wh=1, layers=0, label=None, addtomodel=True):
        self.storeinput(inspect.currentframe())
        LineSinkBase.__init__(self, model, x1, y1, x2, y2, Qls=0, \
                              res=res, wh=wh, layers=layers,
                              name='HeadLineSink', label=label, \
                              addtomodel=addtomodel)
        self.hc = np.atleast_1d(float(hls))
        self.nunknowns = self.nparam

    def initialize(self):
        LineSinkBase.initialize(self)
        self.pc = self.hc * self.aq.T[self.layers]  # Needed in solving

    def setparams(self, sol):
        self.parameters[:, 0] = sol


class LineSinkHoBase(LineSinkChangeTrace, Element):
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, \
                 Qls=0.0, layers=0, order=0, name='LineSinkHoBase', \
                 label=None, addtomodel=True, aq=None, zcinout=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers, \
                         name=name, label=label)
        self.x1 = float(x1)
        self.y1 = float(y1)
        self.x2 = float(x2)
        self.y2 = float(y2)
        self.Qls = np.atleast_1d(Qls)
        self.order = order
        self.nparam = self.nlayers * (self.order + 1)
        self.addtomodel = addtomodel
        if addtomodel: self.model.add_element(self)
        self.aq = aq
        self.zcinout = zcinout
        if self.model.f2py:
            self.bessel = besselaesnew.besselaesnew
        else:
            self.bessel = besselaesnumba
            
    def __repr__(self):
        return self.name + ' from ' + str((self.x1, self.y1)) + ' to ' + str(
               (self.x2, self.y2))

    def initialize(self):
        self.ncp = self.order + 1
        self.z1 = self.x1 + 1j * self.y1
        self.z2 = self.x2 + 1j * self.y2
        self.L = np.abs(self.z1 - self.z2)
        #bchanged minus to plus
        self.theta_norm_out = np.arctan2(self.y2 - self.y1,
                                       self.x2 - self.x1) + np.pi / 2.0
        self.cosnorm = np.cos(self.theta_norm_out) * np.ones(self.ncp)
        self.sinnorm = np.sin(self.theta_norm_out) * np.ones(self.ncp)
        # array of ncp by nlayers * (order + 1)
        self.strengthinf = strengthinf_controlpoints(self.ncp, self.nlayers) 
        #
        self.xc, self.yc = controlpoints(self.ncp, self.z1, self.z2, eps=0)
        if self.zcinout is not None:
            self.xcin, self.ycin = controlpoints(self.ncp, self.zcinout[0],
                                                 self.zcinout[1], eps=0)
            self.xcout, self.ycout = controlpoints(self.ncp, self.zcinout[2],
                                                   self.zcinout[3], eps=0)
        else:
            self.xcin, self.ycin = controlpoints(self.ncp, self.z1, self.z2,
                                                 eps=1e-6)
            self.xcout, self.ycout = controlpoints(self.ncp, self.z1, self.z2,
                                                   eps=-1e-6)
        if self.aq is None:
            self.aq = self.model.aq.find_aquifer_data(self.xc[0], self.yc[0])
        if self.addtomodel:
            self.aq.add_element(self)
        self.parameters = np.empty((self.nparam, 1))
        # Not sure if that needs to be here
        self.parameters[:, 0] = self.Qls / self.L

    def potinf(self, x, y, aq=None):
        '''Can be called with only one x,y value
        Returns array(nparam, self.aq.naq) with order
        order 0, layer[0]
        order 0, layer[1]
        ...
        order 1, layer[0]
        order 1, layer[1]
        etc
        '''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq))
        if aq == self.aq:
            # clever way of using a reshaped rv here
            potrv = rv.reshape((self.order + 1, self.nlayers, aq.naq))
            pot = np.zeros((self.order + 1, aq.naq))
            pot[:, :] = self.bessel.potbeslsv(float(x), float(y), self.z1,
                self.z2, aq.lab, self.order, aq.ilap, aq.naq)
            potrv[:] = self.aq.coef[self.layers] * pot[:, np.newaxis, :]
        return rv

    def disvecinf(self, x, y, aq=None):
        '''Can be called with only one x,y value
        Returns array(nparam, self.aq.naq) with order
        order 0, layer[0]
        order 0, layer[1]
        ...
        order 1, layer[0]
        order 1, layer[1]
        etc
        '''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nparam, aq.naq))
        if aq == self.aq:
            qxqyrv = rv.reshape((2, self.order + 1, self.nlayers, aq.naq))
            qxqy = np.zeros((2 * (self.order + 1), aq.naq))
            qxqy[:, :] = self.bessel.disbeslsv(float(x), float(y), self.z1,
                self.z2, aq.lab, self.order, aq.ilap, aq.naq)
            qxqyrv[0, :] = self.aq.coef[self.layers] * qxqy[:self.order + 1,
                                                       np.newaxis, :]
            qxqyrv[1, :] = self.aq.coef[self.layers] * qxqy[self.order + 1:,
                                                       np.newaxis, :]
        return rv

    def plot(self, layer=None):
        if (layer is None) or (layer in self.layers):
            plt.plot([self.x1, self.x2], [self.y1, self.y2], 'k')

    def dischargeinf(self):
        # returns the unit contribution to the discharge in each layer
        # array of length nunknowns
        Qdisinf = np.zeros((self.order + 1, self.nlayers))
        for n in range(self.order + 1):
            Qdisinf[n] =  (1 ** (n + 1) - (-1) ** (n + 1)) / (n + 1)
        rv = self.L / 2 * Qdisinf.ravel()
        return rv

    def discharge(self):
        # returns the discharge in each layer
        rv = np.zeros(self.aq.naq)
        Qls = self.parameters[:, 0] * self.dischargeinf()
        rv[self.layers] = np.sum(Qls.reshape(self.order + 1, self.nlayers), 0)
        return rv

    def headinside(self, icp=0):
        hinside = \
            self.model.head(self.xc[icp], self.yc[icp])[self.layers[0]] \
            - np.sum(self.strengthinf[icp] * self.parameters[:, 0]) \
            * self.res / self.whfac
        return hinside

class HeadLineSink(LineSinkHoBase, HeadEquation):
    """
    Create a head-specified line-sink
    which may optionally have a width and resistance

    Parameters
    ----------

    model : Model object
        Model to which the element is added
    x1 : scalar
        x-coordinate of fist point of line-sink
    y1 : scalar
        y-coordinate of fist point of line-sink
    x2 : scalar
        x-coordinate of second point of line-sink
    y2 : scalar
        y-coordinate of second point of line-sink
    hls : scalar, array or list
        head along line-sink
        if scalar: head is the same everywhere along line-sink
        if list or array of length 2: head at beginning and end of line-sink
        if list or array with length order + 1: heads at control points
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the aquifer layer
    order : int (default is 0)
        polynomial order or inflow along line-sink
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers
    label: str or None
        label of element

    See Also
    --------

    :class:`.HeadLineSinkString`

    """

    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, \
                 hls=1.0, res=0, wh=1, order=0, layers=0, \
                 label=None, name='HeadLineSink', addtomodel=True):
        self.storeinput(inspect.currentframe())
        LineSinkHoBase.__init__(self, model, x1, y1, x2, y2, Qls=0, \
                                layers=layers, order=order,
                                name=name, label=label, \
                                addtomodel=addtomodel)
        self.hls = np.atleast_1d(hls)
        self.res = res
        self.wh = wh
        self.nunknowns = self.nparam

    def initialize(self):
        LineSinkHoBase.initialize(self)
        if self.wh == 'H':
            self.whfac = self.aq.Haq[self.layers]
        elif self.wh == '2H':
            self.whfac = 2.0 * self.aq.Haq[self.layers]
        elif np.isscalar(self.wh):
            self.whfac = self.wh * np.ones(self.nlayers)
        self.resfac = np.tile(self.res / self.whfac, self.ncp) * self.strengthinf
        self.resfac.shape = (self.ncp, self.nlayers, self.nunknowns)
        if len(self.hls) == 1:
            self.hc = self.hls * np.ones(self.nparam)
        elif len(self.hls) == self.ncp:  # head specified at control points
            self.hc = np.repeat(self.hls, self.nlayers)
        elif len(self.hls) == 2:
            s = np.sqrt((self.xc - self.x1) ** 2 + (self.yc - self.y1) ** 2)
            self.hc = np.interp(s, [0, self.L], self.hls)
            self.hc = np.repeat(self.hc, self.nlayers)

    def setparams(self, sol):
        self.parameters[:, 0] = sol

class LineSinkDitch(HeadLineSink):
    """
    Class to create a line-sink for which the total discharge
    is specified, and for which the head along the line-sink
    is uniform but unknown.

    Parameters
    ----------

    model : Model object
        Model to which the element is added
    x1 : scalar
        x-coordinate of fist point of line-sink
    y1 : scalar
        y-coordinate of fist point of line-sink
    x2 : scalar
        x-coordinate of second point of line-sink
    y2 : scalar
        y-coordinate of second point of line-sink
    Qls : scalar
        total discharge of the line-sink
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the aquifer layer
    order : int (default is 0)
        polynomial order or inflow along line-sink
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers
    label: str or None
        label of element

    See Also
    --------

    :class:`.LineSinkDitchString`

    """

    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, \
                 Qls=1, res=0, wh=1, order=0, layers=0, label=None, addtomodel=True):
        self.storeinput(inspect.currentframe())
        HeadLineSink.__init__(self, model, x1, y1, x2, y2, \
                 hls=0, res=res, wh=wh, order=order, layers=layers, label=label,
                 name='HeadLineSinkDitch', addtomodel=addtomodel)
        self.Qls = Qls

    def initialize(self):
        HeadLineSink.initialize(self)

    def equation(self):
        mat, rhs = HeadLineSink.equation(self)
        for i in range(1, self.nunknowns):
            mat[i] -= mat[0]
            rhs[i] -= rhs[0]
        # first equation is sum of discharges equals Qls
        mat[0] = 0
        ieq = 0
        for e in self.model.elementlist:
            if e.nunknowns > 0:
                if e == self:
                    mat[0, ieq:ieq + e.nunknowns] = self.dischargeinf()
                    break
                ieq += e.nunknowns
        rhs[0] = self.Qls
        return mat, rhs

    def setparams(self, sol):
        self.parameters[:, 0] = sol


class LineSinkStringBase(Element):
    '''Original implementation
    Used for boundaries of inhomogenieities'''
    def __init__(self, model, xy, closed=False, layers=0, order=0,
                 name='LineSinkStringBase', label=None, aq=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers, \
                         name=name, label=label)
        self.xy = np.atleast_2d(xy).astype('d')
        if closed: self.xy = np.vstack((self.xy, self.xy[0]))
        self.order = order
        self.aq = aq
        self.lslist = []
        self.x, self.y = self.xy[:, 0], self.xy[:, 1]
        self.nls = len(self.x) - 1
        for i in range(self.nls):
            if label is not None:
                lslabel = label + '_' + str(i)
            else:
                lslabel = label
            self.lslist.append(LineSinkHoBase(model,
                x1=self.x[i], y1=self.y[i], 
                x2=self.x[i + 1], y2=self.y[i + 1],
                Qls=0.0, layers=layers, order=order, label=lslabel,
                addtomodel=False, aq=aq))

    def __repr__(self):
        return self.name + ' with nodes ' + str(self.xy)

    def initialize(self):
        for ls in self.lslist:
            ls.initialize()
        # Same order for all elements in string
        self.ncp = self.nls * self.lslist[0].ncp
        self.nparam = self.nls * self.lslist[0].nparam
        self.nunknowns = self.nparam
        self.xls = np.empty((self.nls, 2))
        self.yls = np.empty((self.nls, 2))
        for i, ls in enumerate(self.lslist):
            self.xls[i, :] = [ls.x1, ls.x2]
            self.yls[i, :] = [ls.y1, ls.y2]
        if self.aq is None:
            self.aq = self.model.aq.find_aquifer_data(self.lslist[0].xc,
                                                      self.lslist[0].yc)
        self.parameters = np.zeros((self.nparam, 1))
        # As parameters are only stored for the element not the list,
        # we need to combine the following
        self.xc = np.array([ls.xc for ls in self.lslist]).flatten()
        self.yc = np.array([ls.yc for ls in self.lslist]).flatten()
        self.xcin = np.array([ls.xcin for ls in self.lslist]).flatten()
        self.ycin = np.array([ls.ycin for ls in self.lslist]).flatten()
        self.xcout = np.array([ls.xcout for ls in self.lslist]).flatten()
        self.ycout = np.array([ls.ycout for ls in self.lslist]).flatten()
        self.cosnorm = np.array([ls.cosnorm for ls in self.lslist]).flatten()
        self.sinnorm = np.array([ls.sinnorm for ls in self.lslist]).flatten()
        self.aqin = self.model.aq.find_aquifer_data(self.xcin[0], self.ycin[0])
        self.aqout = self.model.aq.find_aquifer_data(self.xcout[0],
                                                     self.ycout[0])

    def potinf(self, x, y, aq=None):
        '''
        linesink 0, order 0, layer[0]
                    order 0, layer[1]
                    ...
                    order 1, layer[0]
                    order 1, layer[1]
                    ...
        linesink 1, order 0, layer[0]
                    order 0, layer[1]
                    ...
                    order 1, layer[0]
                    order 1, layer[1]
                    ...
        '''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nls, self.lslist[0].nparam, aq.naq))
        for i in range(self.nls):
            rv[i] = self.lslist[i].potinf(x, y, aq)
        rv.shape = (self.nparam, aq.naq)
        return rv

    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nls, self.lslist[0].nparam, aq.naq))
        for i in range(self.nls):
            rv[:, i] = self.lslist[i].disvecinf(x, y, aq)
        rv.shape = (2, self.nparam, aq.naq)
        return rv

    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer,
                    direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        for ls in self.lslist:
            changed, terminate, xyztnew, message = ls.changetrace(
                xyzt1, xyzt2, aq, layer, ltype, modellayer, direction)
            if changed or terminate:
                return changed, terminate, xyztnew, message
        return changed, terminate, xyztnew, message

    def plot(self, layer=None):
        if (layer is None) or (layer in self.layers):
            plt.plot(self.x, self.y, 'k')

class HeadLineSinkStringOLd(LineSinkStringBase, HeadEquation):
    def __init__(self, model, xy=[(-1, 0), (1, 0)], hls=0.0, \
                 layers=0, order=0, label=None):
        self.storeinput(inspect.currentframe())
        LineSinkStringBase.__init__(self, model, xy, closed=False,
                                    layers=layers, order=order, \
                                    name='HeadLineSinkString', label=label,
                                    aq=None)
        self.hls = np.atleast_1d(hls)
        self.model.add_element(self)

    def initialize(self):
        LineSinkStringBase.initialize(self)
        self.aq.add_element(self)
        # self.pc = np.array([ls.pc for ls in self.lslist]).flatten()
        if len(self.hls) == 1:
            self.pc = self.hls * self.aq.T[self.layers] * np.ones(self.nparam)
        elif len(self.hls) == self.nls:  # head specified at centers
            self.pc = (self.hls[:, np.newaxis] *
                       self.aq.T[self.layers]).flatten()
        elif len(self.hls) == 2:
            L = np.array([ls.L for ls in self.lslist])
            Ltot = np.sum(L)
            xp = np.zeros(self.nls)
            xp[0] = 0.5 * L[0]
            for i in range(1, self.nls):
                xp[i] = xp[i - 1] + 0.5 * (L[i - 1] + L[i])
            self.hls = np.interp(xp, [0, Ltot], self.hls)
            self.pc = (self.hls[:, np.newaxis] *
                       self.aq.T[self.layers]).flatten()
        else:
            print('Error: hls entry not supported')
        self.resfac = 0.0

    def setparams(self, sol):
        self.parameters[:, 0] = sol

class LineSinkStringBase2(Element):
    '''
    Alternative implementation that loops through line-sinks to build equation
    Has the advantage that it is easier to have different line-sinks in
    different layers and/or aquifers
    '''
    def __init__(self, model, xy, closed=False, layers=0, order=0,
                 name='LineSinkStringBase', label=None, aq=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers, \
                         name=name, label=label)
        self.xy = np.atleast_2d(xy).astype('d')
        if closed: self.xy = np.vstack((self.xy, self.xy[0]))
        self.order = order  # same for all segments in string
        self.lslist = []
        self.x, self.y = self.xy[:, 0], self.xy[:, 1]
        self.nls = len(self.x) - 1
        if self.layers.ndim == 1:
            if len(self.layers) == self.nls:
                self.layers = self.layers[:, np.newaxis]
            else:  # entire string in these layers
                self.layers = self.layers * \
                    np.ones((self.nls, len(self.layers)), dtype='int')
        self.nlayers = len(self.layers[0])

    def __repr__(self):
        return self.name + ' with nodes ' + str(self.xy)

    def initialize(self):
        for ls in self.lslist:
            ls.initialize()
        self.aq = []
        for ls in self.lslist:
            if ls.aq not in self.aq:
                self.aq.append(ls.aq)
        for aq in self.aq:
            aq.add_element(self)
        # Same order for all elements in string
        #self.ncp = sum(ls.ncp for ls in self.lslist)
        self.nparam = sum(ls.nparam for ls in self.lslist)
        self.nunknowns = self.nparam
        # where are self.xls and self.yls used? self.xls and self.yls removed
        self.parameters = np.zeros((self.nparam, 1))

    def potinf(self, x, y, aq=None):
        '''
        linesink 0, order 0, layer[0]
                    order 0, layer[1]
                    ...
                    order 1, layer[0]
                    order 1, layer[1]
                    ...
        linesink 1, order 0, layer[0]
                    order 0, layer[1]
                    ...
                    order 1, layer[0]
                    order 1, layer[1]
                    ...
        '''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nls, self.lslist[0].nparam, aq.naq))
        if aq in self.aq:
            for i, ls in enumerate(self.lslist):
                rv[i] = ls.potinf(x, y, aq)
        rv.shape = (self.nparam, aq.naq)
        return rv

    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nls, self.lslist[0].nparam, aq.naq))
        if aq in self.aq:
            for i, ls in enumerate(self.lslist):
                rv[:, i] = ls.disvecinf(x, y, aq)
        rv.shape = (2, self.nparam, aq.naq)
        return rv

    def dischargeinf(self):
        rv = np.zeros((self.nls, self.lslist[0].nparam))
        for i, ls in enumerate(self.lslist):
            rv[i] = ls.dischargeinf()
        return rv.ravel()

    def discharge(self):
        """Discharge of the element in each layer
        """

        rv = np.zeros(self.aq[0].naq)
        Qls = self.parameters[:, 0] * self.dischargeinf()
        Qls.shape = (self.nls, self.nlayers, self.order + 1)
        Qls = np.sum(Qls, 2)
        for i, q in enumerate(Qls):
            rv[self.layers[i]] += q
            #rv[self.layers] = np.sum(Qls.reshape(self.nls * (self.order + 1), self.nlayers), 0)
        return rv

    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer,
                    direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        for ls in self.lslist:
            changed, terminate, xyztnew, message = ls.changetrace(
                xyzt1, xyzt2, aq, layer, ltype, modellayer, direction,
                hstepmax)
            if changed or terminate:
                return changed, terminate, xyztnew, message
        return changed, terminate, xyztnew, message

    def plot(self, layer=None):
        if (layer is None) or (layer in self.layers):
            plt.plot(self.x, self.y, 'k')

class HeadLineSinkString(LineSinkStringBase2):
    """
    Class to create a string of head-specified line-sinks
    which may optionally have a width and resistance

    Parameters
    ----------

    model : Model object
        Model to which the element is added
    xy : array or list
        list or array of (x,y) pairs of coordinates of end-points of
        line-sinks in string
    hls : scalar, array or list
        head along string
        if scalar: head is the same everywhere along the string
        if list or array of length 2: head at beginning and end of string
        if list or array with same length as xy: heads at nodes, which
        may contain nans, except for first and last point
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer 
        layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when 
        flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the 
        aquifer layer
    order : int (default is 0)
        order of all line-sinks in string
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers
    label: str or None

    See Also
    --------

    :class:`.HeadLineSink`

    """

    def __init__(self, model, xy=[(-1, 0), (1, 0)], hls=0,
                 res=0, wh=1, order=0, layers=0, label=None, 
                 name='HeadLineSinkString'):
        self.storeinput(inspect.currentframe())
        LineSinkStringBase2.__init__(self, model, xy, closed=False,
                                    layers=layers, order=order,
                                    name=name, label=label,
                                    aq=None)
        self.hls = np.atleast_1d(hls)
        self.res = res
        self.wh = wh
        self.model.add_element(self)
        #TO DO: TEST FOR DIFFERENT AQUIFERS AND LAYERS

    def initialize(self):
        if len(self.hls) == 1:  # one value
            self.hls = self.hls * np.ones(self.nls + 1)  # at all nodes
        elif len(self.hls) == 2:  # values at beginning and end
            L = np.sqrt((self.x[1:] - self.x[:-1]) ** 2 
                        + (self.y[1:] - self.y[:-1]) ** 2)
            s = np.hstack((0, np.cumsum(L)))
            self.hls = np.interp(s, [0, s[-1]], self.hls)
        elif len(self.hls) == len(self.x):  # nodes may contain nan values
            if np.isnan(self.hls).any():
                L = np.sqrt((self.x[1:] - self.x[:-1]) ** 2 
                            + (self.y[1:] - self.y[:-1]) ** 2)
                s = np.hstack((0, np.cumsum(L)))
                self.hls = np.interp(s, s[~np.isnan(self.hls)],
                                     self.hls[~np.isnan(self.hls)])
        else:
            print('Error: hls entry not supported in HeadLineSinkString')
        self.lslist = []  # start with empty list
        for i in range(self.nls):
            if self.label is not None:
                lslabel = self.label + '_' + str(i)
            else:
                lslabel = self.label
            self.lslist.append(
                HeadLineSink(self.model, x1=self.x[i], y1=self.y[i], 
                             x2=self.x[i + 1], y2=self.y[i + 1], 
                             hls=self.hls[i:i + 2], res=self.res, 
                             wh=self.wh, layers=self.layers[i],
                             order=self.order, label=lslabel, addtomodel=False))
        LineSinkStringBase2.initialize(self)

    def setparams(self, sol):
        self.parameters[:, 0] = sol

    def equation(self):
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.empty(self.nunknowns)
        ieq = 0
        for ls in self.lslist:
            matls, rhsls = ls.equation()
            neq = len(rhsls)
            mat[ieq:ieq + neq] = matls
            rhs[ieq:ieq + neq] = rhsls
            ieq += neq
        # fix to include resistance
        # this is not pretty but works
        # not sure how to change the design to make this nicer
        # I guess the additional matrix can be pre-computed and stored
        jcol = 0
        for e in self.model.elementlist:
            if e == self:
                break
            elif e.nunknowns > 0:
                jcol += e.nunknowns
        irow = 0
        for ls in self.lslist:
            for icp in range(ls.ncp):
                mat[irow:irow+ ls.nlayers, jcol:jcol + ls.nunknowns] -= ls.resfac[icp]
                irow += ls.nlayers
            jcol += ls.nunknowns
        return mat, rhs

class LineSinkDitchString(HeadLineSinkString):
    """
    Class to create a string of LineSinkDitch elements for which the
    total discharge of the string is specified, and for which the head
    along the entire string is uniform but unknown.

    Parameters
    ----------

    model : Model object
        Model to which the element is added
    xy : array or list
        list or array of (x,y) pairs of coordinates of end-points of
        line-sinks in string
    Qls : scalar
        total discharge of the string
    res : scalar (default is 0)
        resistance of line-sinks in string
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the aquifer layer
    order : int (default is 0)
        polynomial order or inflow along each line-sink in string
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers
    label: str or None
        label of element

    See Also
    --------

    :class:`.LineSinkDitch`

    """

    def __init__(self, model, xy=[(-1, 0), (1, 0)], \
                 Qls=1, res=0, wh=1, order=0, layers=0, label=None):
        self.storeinput(inspect.currentframe())
        HeadLineSinkString.__init__(self, model, xy=xy, hls=0, \
                 res=res, wh=wh, order=order, layers=layers, label=label,
                 name='LineSinkDitchString')
        self.Qls = Qls

    def initialize(self):
        HeadLineSinkString.initialize(self)

    def equation(self):
        mat, rhs = HeadLineSinkString.equation(self)
        for i in range(1, self.nunknowns):
            mat[i] -= mat[0]
            rhs[i] -= rhs[0]
        # first equation is sum of discharges equals Qls
        mat[0] = 0
        ieq = 0
        for e in self.model.elementlist:
            if e.nunknowns > 0:
                if e == self:
                    mat[0, ieq:ieq + self.nunknowns] = self.dischargeinf()
                    break
                ieq += e.nunknowns
        rhs[0] = self.Qls
        return mat, rhs

    def setparams(self, sol):
        self.parameters[:, 0] = sol
        
class LineSinkContainer(Element):
    '''
    Container class for bunch of line-sinks
    Required attributes:
    lslist: list of line-sinks
    nls: total number of line-sinks
    '''
    def __init__(self, model, layers=0, order=0,
                 name='LineSinkStringContainer', label=None, aq=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers, \
                         name=name, label=label)
        self.order = order
#         self.xy = np.atleast_2d(xy).astype('d')
#         if closed: self.xy = np.vstack((self.xy, self.xy[0]))
#         self.order = order  # same for all segments in string
#         self.lslist = []
#         self.x, self.y = self.xy[:, 0], self.xy[:, 1]
#         self.nls = len(self.x) - 1
#         if self.layers.ndim == 1:
#             if len(self.layers) == self.nls:
#                 self.layers = self.layers[:, np.newaxis]
#             else:  # entire string in these layers
#                 self.layers = self.layers * \
#                     np.ones((self.nls, len(self.layers)), dtype='int')
#         self.nlayers = len(self.layers[0])

    def __repr__(self):
        return self.name

    def initialize(self):
        self.layers = []
        for ls in self.lslist:
            ls.initialize()
            if ls.layers[0] not in self.layers:  # works if ls in 1 layer
                self.layers.append(ls.layers[0])
        self.aq = []
        for ls in self.lslist:
            if ls.aq not in self.aq:
                self.aq.append(ls.aq)
        for aq in self.aq:
            aq.add_element(self)
        self.nparam = sum(ls.nparam for ls in self.lslist)
        self.nunknowns = self.nparam
        self.parameters = np.zeros((self.nparam, 1))

    def potinf(self, x, y, aq=None):
        '''
        linesink 0, order 0, layer[0]
                    order 0, layer[1]
                    ...
                    order 1, layer[0]
                    order 1, layer[1]
                    ...
        linesink 1, order 0, layer[0]
                    order 0, layer[1]
                    ...
                    order 1, layer[0]
                    order 1, layer[1]
                    ...
        '''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nls, self.lslist[0].nparam, aq.naq))
        if aq in self.aq:
            for i, ls in enumerate(self.lslist):
                rv[i] = ls.potinf(x, y, aq)
        rv.shape = (self.nparam, aq.naq)
        return rv

    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nls, self.lslist[0].nparam, aq.naq))
        if aq in self.aq:
            for i, ls in enumerate(self.lslist):
                rv[:, i] = ls.disvecinf(x, y, aq)
        rv.shape = (2, self.nparam, aq.naq)
        return rv

    def dischargeinf(self):
        rv = np.zeros((self.nls, self.lslist[0].nparam))
        for i, ls in enumerate(self.lslist):
            rv[i] = ls.dischargeinf()
        return rv.ravel()

    def discharge(self):
        """Discharge of the element in each layer
        """
        # TBD: it is unclear what this function does. shouldn't it return
        # values only for the layers where the element is screened?
        rv = np.zeros(self.aq[0].naq)
        Qls = self.parameters[:, 0] * self.dischargeinf()
        Qls.shape = (self.nls, self.nlayers, self.order + 1)
        Qls = np.sum(Qls, 2)
        for i, q in enumerate(Qls):
            rv[self.layers[i]] += q
            #rv[self.layers] = np.sum(Qls.reshape(self.nls * (self.order + 1), self.nlayers), 0)
        return rv

    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer,
                    direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        for ls in self.lslist:
            changed, terminate, xyztnew, message = ls.changetrace(
                xyzt1, xyzt2, aq, layer, ltype, modellayer, direction,
                hstepmax)
            if changed or terminate:
                return changed, terminate, xyztnew, message
        return changed, terminate, xyztnew, message

    def plot(self, layer=None):
        if (layer is None) or (layer in self.layers):
            for i in range(len(self.xls)):
                plt.plot(self.xls[i], self.yls[i], 'k')
            
class HeadLineSinkContainer(LineSinkContainer):
    """
    Class to create a container of head-specified line-sinks
    which may optionally have a width and resistance

    Parameters
    ----------

    model : Model object
        Model to which the element is added
    xydict : dictionary
        dictionary with lists or arrays of (x,y) pairs of coordinates of 
        end-points of line-sinks that are stringed together
    hls : scalar
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer 
        layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when 
        flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the 
        aquifer layer
    order : int (default is 0)
        order of all line-sinks in string
    laydict : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers
    label: str or None

    See Also
    --------

    :class:`.HeadLineSink`

    """

    def __init__(self, model, xydict={0: [(-1, 0), (1, 0)]}, hls=0,
                 res=0, wh=1, order=0, laydict={0: 0}, label=None, 
                 name='HeadLineSinkContainer'):
        self.storeinput(inspect.currentframe())
        LineSinkContainer.__init__(self, model, layers=0, 
                                         order=order, name=name, label=label,
                                         aq=None)
        self.xydict = xydict
        self.hls = hls
        self.laydict = laydict
        self.res = res
        self.wh = wh
        self.model.add_element(self)
        #TO DO: TEST FOR DIFFERENT AQUIFERS AND LAYERS

    def initialize(self):
        self.lslist = []
        self.xls = [] # used for layout
        self.yls = []
        for key in self.xydict.keys():
            xy = np.atleast_1d(self.xydict[key])
            layers = np.atleast_1d(self.laydict[key])
            if len(layers) == 1:
                layers = layers * np.ones(len(xy) - 1, dtype='int')
            self.xls.append(xy[:, 0])
            self.yls.append(xy[:, 1]) 
            for i in range(len(xy) - 1):
                x1, y1 = xy[i]
                x2, y2 = xy[i + 1]
                ls = HeadLineSink(self.model, x1=x1, y1=y1, x2=x2, y2=y2, 
                                  hls=self.hls, res=self.res, wh=self.wh, 
                                  layers=layers[i], order=self.order, 
                                  label=None, addtomodel=False)
                self.lslist.append(ls)
        self.nls = len(self.lslist)
        for i in range(self.nls):
            if self.label is not None:
                lslabel = self.label + '_' + str(i)
            else:
                lslabel = self.label
        LineSinkContainer.initialize(self)

    def setparams(self, sol):
        self.parameters[:, 0] = sol

    def equation(self):
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.empty(self.nunknowns)
        ieq = 0
        for ls in self.lslist:
            matls, rhsls = ls.equation()
            neq = len(rhsls)
            mat[ieq:ieq + neq] = matls
            rhs[ieq:ieq + neq] = rhsls
            ieq += neq
        # fix to include resistance
        # this is not pretty but works
        # not sure how to change the design to make this nicer
        # I guess the additional matrix can be pre-computed and stored
        jcol = 0
        for e in self.model.elementlist:
            if e == self:
                break
            elif e.nunknowns > 0:
                jcol += e.nunknowns
        irow = 0
        for ls in self.lslist:
            for icp in range(ls.ncp):
                mat[irow:irow+ ls.nlayers, jcol:jcol + ls.nunknowns] -= ls.resfac[icp]
                irow += ls.nlayers
            jcol += ls.nunknowns
        return mat, rhs



import numpy as np
import inspect  # Used for storing the input
from .aquifer import AquiferData
from .aquifer_parameters import param_maq
from .constant import ConstantInside, ConstantStar
from .intlinesink import IntHeadDiffLineSink, IntFluxDiffLineSink, IntFluxLineSink
from .controlpoints import controlpoints

__all__ = ['PolygonInhomMaq']

class PolygonInhom(AquiferData):
    tiny = 1e-8

    def __init__(self, model, xy, kaq, c, z, npor, ltype, hstar,
                 order, ndeg):
        # All input variables except model should be numpy arrays
        # That should be checked outside this function):        
        AquiferData.__init__(self, model, kaq, c, z, npor, ltype)
        self.order = order
        self.ndeg = ndeg
        self.hstar = hstar
        self.inhom_number = self.model.aq.add_inhom(self)
        self.z1, self.z2 = compute_z1z2(xy)
        self.Nsides = len(self.z1)
        Zin = 1e-6j
        Zout = -1e-6j
        self.zcin = 0.5 * (self.z2 - self.z1) * Zin + 0.5 * (
        self.z1 + self.z2)  # point at center on inside
        self.zcout = 0.5 * (self.z2 - self.z1) * Zout + 0.5 * (
        self.z1 + self.z2)  # point at center on outside
        self.x = np.hstack((self.z1.real, self.z2[-1].real))
        self.y = np.hstack((self.z1.imag, self.z2[-1].imag))
        self.xmin = min(self.x)
        self.xmax = max(self.x)
        self.ymin = min(self.y)
        self.ymax = max(self.y)

    def __repr__(self):
        return 'PolygonInhom: ' + str(list(zip(self.x, self.y)))

    def isinside(self, x, y):
        rv = 0
        if (x >= self.xmin) and (x <= self.xmax) and (y >= self.ymin) and (
            y <= self.ymax):
            z = complex(x, y)
            bigZ = (2.0 * z - (self.z1 + self.z2)) / (self.z2 - self.z1)
            bigZmin1 = bigZ - 1.0
            bigZplus1 = bigZ + 1.0
            minAbsBigZmin1 = min(abs(bigZmin1))
            minAbsBigZplus1 = min(abs(bigZplus1))
            if minAbsBigZmin1 < self.tiny or minAbsBigZplus1 < self.tiny:
                rv = 1
                return rv
            angles = np.log(bigZmin1 / bigZplus1).imag
            angle = np.sum(angles)
            if angle > np.pi: rv = 1
        return rv

    def create_elements(self):
        aqin = self.model.aq.find_aquifer_data(self.zcin[0].real,
                                               self.zcin[0].imag)
        for i in range(self.Nsides):
            aqout = self.model.aq.find_aquifer_data(self.zcout[i].real,
                                                    self.zcout[i].imag)
            if (aqout == self.model.aq) or (
                aqout.inhom_number > self.inhom_number):
                ls = IntHeadDiffLineSink(self.model, x1=self.x[i], y1=self.y[i],
                                         x2=self.x[i + 1], y2=self.y[i + 1], \
                                         order=self.order, ndeg=self.ndeg,
                                         label=None, addtomodel=True, \
                                         aq=aqin, aqin=aqin, aqout=aqout)
                ls = IntFluxDiffLineSink(self.model, x1=self.x[i], y1=self.y[i],
                                         x2=self.x[i + 1], y2=self.y[i + 1], \
                                         order=self.order, ndeg=self.ndeg,
                                         label=None, addtomodel=True, \
                                         aq=aqout, aqin=aqin, aqout=aqout)
        if aqin.ltype[0] == 'a':  # add constant on inside
            c = ConstantInside(self.model, self.zcin.real, self.zcin.imag)
            c.inhomelement = True
        if aqin.ltype[0] == 'l':
            assert self.hstar is not None, 'Error: hstar needs to be set'
            c = ConstantStar(self.model, self.hstar, aq=aqin)
            c.inhomelement = True


class PolygonInhomMaq(PolygonInhom):
    """Create a polygonal inhomogeneity
    
    Parameters
    ----------
    model : Model object
        model to which the element is added
    xy : array or list
        list or array of (x,y) pairs of coordinates of corners of the
        inhomogeneity
        polygonal boundary is automatically closed (so first point
        is not repeated)
    kaq : float, array or list
        hydraulic conductivity of each aquifer from the top down
        if float, hydraulic conductivity is the same in all aquifers
    z : array or list
        elevation tops and bottoms of the aquifers from the top down
        leaky layers may have zero thickness
        if topboundary='conf': length is 2 * number of aquifers
        if topboundary='semi': length is 2 * number of aquifers + 1 as top
        of leaky layer on top of systems needs to be specified
    c : float, array or list
        resistance of leaky layers from the top down
        if float, resistance is the same for all leaky layers
        if topboundary='conf': length is number of aquifers - 1
        if topboundary='semi': length is number of aquifers
    npor : float, array or list
        porosity of all aquifers and leaky layers from the top down
        if float, porosity is the same for all layers
        if topboundary='conf': length is 2 * number of aquifers - 1
        if topboundary='semi': length is 2 * number of aquifers
    topboundary : string, 'conf' or 'semi' (default is 'conf')
        indicating whether the top is confined ('conf') or
        semi-confined ('semi')
    hstar : float or None (default is None)
        head value above semi-confining top, only read if topboundary='semi'
    order : int
        polynomial order of flux along each segment
    ndeg : int
        number of points used between two segments to numerically
        integrate normal discharge

    """
    
    tiny = 1e-8

    def __init__(self, model, xy, kaq=1, z=[1, 0], c=[], npor=0.3, topboundary='conf',
                 hstar=None, order=3, ndeg=3):
        self.storeinput(inspect.currentframe())
        kaq, c, npor, ltype, = param_maq(kaq, z, c, npor, topboundary)
        PolygonInhom.__init__(self, model, xy, kaq, c, z, npor, ltype,
                              hstar, order, ndeg)


def compute_z1z2(xy):
    # Returns z1 and z2 of polygon, in clockwise order
    x, y = list(zip(*xy))
    if x[0] == x[-1] and y[0] == y[-1]:  # In case last point is repeated
        x = x[:-1];
        y = y[:-1]
    z1 = np.array(x) + np.array(y) * 1j
    index = list(range(1, len(z1))) + [0]
    z2 = z1[index]
    Z = 1e-6j
    z = Z * (z2[0] - z1[0]) / 2.0 + 0.5 * (z1[0] + z2[0])
    bigZ = (2.0 * z - (z1 + z2)) / (z2 - z1)
    bigZmin1 = bigZ - 1.0
    bigZplus1 = bigZ + 1.0
    angle = np.sum(np.log(bigZmin1 / bigZplus1).imag)
    if angle < np.pi:  # reverse order
        z1 = z1[::-1]
        z2 = z1[index]
    return z1, z2


class BuildingPit(AquiferData):
    tiny = 1e-8

    def __init__(self, model, xy, kaq=1, z=[1, 0], c=[], npor=0.3,
                 topboundary="conf", hstar=None, order=3,
                 ndeg=3, layers=[0]):
        """Element to simulate a building pit surrounded by an
        impermeable wall. Layers with wall are provided with layers
        argument.

        Parameters
        ----------
        model : Model object
            model to which the element is added
        xy : array or list
            list or array of (x,y) pairs of coordinates of corners of the
            inhomogeneity
            polygonal boundary is automatically closed (so first point
            is not repeated)
        kaq : float, array or list
            hydraulic conductivity of each aquifer from the top down
            if float, hydraulic conductivity is the same in all aquifers
        z : array or list
            elevation tops and bottoms of the aquifers from the top down
            leaky layers may have zero thickness
            if topboundary='conf': length is 2 * number of aquifers
            if topboundary='semi': length is 2 * number of aquifers + 1 as top
            of leaky layer on top of systems needs to be specified
        c : float, array or list
            resistance of leaky layers from the top down
            if float, resistance is the same for all leaky layers
            if topboundary='conf': length is number of aquifers - 1
            if topboundary='semi': length is number of aquifers
        npor : float, array or list
            porosity of all aquifers and leaky layers from the top down
            if float, porosity is the same for all layers
            if topboundary='conf': length is 2 * number of aquifers - 1
            if topboundary='semi': length is 2 * number of aquifers
        topboundary : string, 'conf' or 'semi' (default is 'conf')
            indicating whether the top is confined ('conf') or
            semi-confined ('semi'). For a building pit, the 'conf'
            option is generally more applicable.
        hstar : float or None (default is None)
            head value above semi-confining top, only read if topboundary='semi'
        order : int
            polynomial order of flux along each segment
        ndeg : int
            number of points used between two segments to numerically
            integrate normal discharge
        layers: list or np.array
            layers in which impermeable wall is present.

        """
        # All input variables except model should be numpy arrays
        # That should be checked outside this function):
        kaq, c, npor, ltype, = param_maq(kaq, z, c, npor, topboundary)
        AquiferData.__init__(self, model, kaq, c, z, npor, ltype)
        self.order = order
        self.ndeg = ndeg

        self.layers = layers  # layers with impermeable wall
        self.nonimplayers = list(set(range(self.model.aq.naq)) - set(self.layers))  # layers without wall

        self.hstar = hstar

        self.inhom_number = self.model.aq.add_inhom(self)
        self.z1, self.z2 = compute_z1z2(xy)
        self.Nsides = len(self.z1)
        Zin = 1e-6j
        Zout = -1e-6j
        self.zcin = 0.5 * (self.z2 - self.z1) * Zin + 0.5 * (
        self.z1 + self.z2)  # point at center on inside
        self.zcout = 0.5 * (self.z2 - self.z1) * Zout + 0.5 * (
        self.z1 + self.z2)  # point at center on outside
        self.x = np.hstack((self.z1.real, self.z2[-1].real))
        self.y = np.hstack((self.z1.imag, self.z2[-1].imag))
        self.xmin = min(self.x)
        self.xmax = max(self.x)
        self.ymin = min(self.y)
        self.ymax = max(self.y)

    def isinside(self, x, y):
        rv = 0
        if (x >= self.xmin) and (x <= self.xmax) and (y >= self.ymin) and (
            y <= self.ymax):
            z = complex(x, y)
            bigZ = (2.0 * z - (self.z1 + self.z2)) / (self.z2 - self.z1)
            bigZmin1 = bigZ - 1.0
            bigZplus1 = bigZ + 1.0
            minAbsBigZmin1 = min(abs(bigZmin1))
            minAbsBigZplus1 = min(abs(bigZplus1))
            if minAbsBigZmin1 < self.tiny or minAbsBigZplus1 < self.tiny:
                rv = 1
                return rv
            angles = np.log(bigZmin1 / bigZplus1).imag
            angle = np.sum(angles)
            if angle > np.pi: rv = 1
        return rv

    def create_elements(self):
        aqin = self.model.aq.find_aquifer_data(self.zcin[0].real,
                                               self.zcin[0].imag)
        for i in range(self.Nsides):
            aqout = self.model.aq.find_aquifer_data(self.zcout[i].real,
                                                    self.zcout[i].imag)
            if (aqout == self.model.aq) or (
                aqout.inhom_number > self.inhom_number):

                # Conditions for layers with impermeable walls
                IntFluxLineSink(self.model, x1=self.x[i], y1=self.y[i],
                                x2=self.x[i + 1], y2=self.y[i + 1],
                                layers=self.layers,
                                order=self.order, ndeg=self.ndeg,
                                label=None, addtomodel=True,
                                aq=aqin, aqin=aqin, aqout=aqout)
                IntFluxLineSink(self.model, x1=self.x[i], y1=self.y[i],
                                x2=self.x[i + 1], y2=self.y[i + 1],
                                layers=self.layers,
                                order=self.order, ndeg=self.ndeg,
                                label=None, addtomodel=True,
                                aq=aqout, aqin=aqin, aqout=aqout)

                # use these conditions for layers without impermeable walls
                IntHeadDiffLineSink(self.model, x1=self.x[i], y1=self.y[i],
                                    x2=self.x[i + 1], y2=self.y[i + 1],
                                    layers=self.nonimplayers,
                                    order=self.order, ndeg=self.ndeg,
                                    label=None, addtomodel=True,
                                    aq=aqin, aqin=aqin, aqout=aqout)
                IntFluxDiffLineSink(self.model, x1=self.x[i], y1=self.y[i],
                                    x2=self.x[i + 1], y2=self.y[i + 1],
                                    layers=self.nonimplayers,
                                    order=self.order, ndeg=self.ndeg,
                                    label=None, addtomodel=True,
                                    aq=aqout, aqin=aqin, aqout=aqout)

        if aqin.ltype[0] == 'a':  # add constant on inside
            c = ConstantInside(self.model, self.zcin.real, self.zcin.imag)
            c.inhomelement = True
        if aqin.ltype[0] == 'l':
            assert self.hstar is not None, 'Error: hstar needs to be set'
            c = ConstantStar(self.model, self.hstar, aq=aqin)
            c.inhomelement = True



import numpy as np
import inspect  # user for storing the input
from .aquifer import AquiferData
from .aquifer_parameters import param_maq, param_3d
from .constant import ConstantStar
from .linesink1d import HeadDiffLineSink1D, FluxDiffLineSink1D
from .stripareasink import StripAreaSinkInhom

__all__ = ['StripInhomMaq', 'StripInhom3D']

class StripInhom(AquiferData):
    tiny = 1e-12
    
    def __init__(self, model, x1, x2, kaq, c, z, npor, ltype, hstar, N):
        AquiferData.__init__(self, model, kaq, c, z, npor, ltype)
        self.x1 = x1
        self.x2 = x2
        self.hstar = hstar
        self.N = N
        self.inhom_number = self.model.aq.add_inhom(self)
        self.addlinesinks = True  # Set to False not to add line-sinks
        
    def __repr__(self):
        return "Inhom1D: " + str(list([self.x1, self.x2]))
        
    def isinside(self, x, y):
        return (x >= self.x1) and (x < self.x2)

    def create_elements(self):
        # HeadDiff on right side, FluxDiff on left side
        if self.x1 == -np.inf:
            xin = self.x2 - self.tiny * abs(self.x2) - self.tiny
            xoutright = self.x2 + self.tiny * abs(self.x2) + self.tiny
            aqin = self.model.aq.find_aquifer_data(xin, 0)
            aqoutright = self.model.aq.find_aquifer_data(xoutright, 0)
            if self.addlinesinks:
                HeadDiffLineSink1D(self.model, self.x2, label=None, 
                                   aq=aqin, aqin=aqin, aqout=aqoutright)
        elif self.x2 == np.inf:
            xin = self.x1 + self.tiny * abs(self.x1) + self.tiny
            xoutleft = self.x1 - self.tiny * abs(self.x1) - self.tiny
            aqin = self.model.aq.find_aquifer_data(xin, 0)
            aqoutleft = self.model.aq.find_aquifer_data(xoutleft, 0)
            if self.addlinesinks:   
                FluxDiffLineSink1D(self.model, self.x1, label=None, 
                                   aq=aqin, aqin=aqin, aqout=aqoutleft)
        else:
            xin = 0.5 * (self.x1 + self.x2)
            xoutleft = self.x1 - self.tiny * abs(self.x1) - self.tiny
            xoutright = self.x2 + self.tiny * abs(self.x2) + self.tiny
            aqin = self.model.aq.find_aquifer_data(xin, 0)
            aqleft = self.model.aq.find_aquifer_data(xoutleft, 0)
            aqright = self.model.aq.find_aquifer_data(xoutright, 0)
            if self.addlinesinks:
                HeadDiffLineSink1D(self.model, self.x2, label=None, 
                                   aq=aqin, aqin=aqin, aqout=aqright)
                FluxDiffLineSink1D(self.model, self.x1, label=None, 
                                   aq=aqin, aqin=aqin, aqout=aqleft)
            if self.N is not None:
                assert aqin.ilap, "Error: infiltration can only be added if topboundary='conf'"
                StripAreaSinkInhom(self.model, self.x1, self.x2, self.N, layer=0)
        if aqin.ltype[0] == 'l':
            assert self.hstar is not None, "Error: hstar needs to be set"
            c = ConstantStar(self.model, self.hstar, aq=aqin)
            c.inhomelement = True
        

class StripInhomMaq(StripInhom):
    """Create a strip inhomogeneity for a mult-aquifer sequence of
    aquifer-leakylayer-aquifer-leakylayer-aquifer etc
    
    Parameters
    ----------
    model : Model object
        model to which the element is added
    x1 : float
        left boundary of inhomogeneity (may be -np.inf if extends to infinity)
    x2 : float
        right boundary of inhomogeneity (may be np.inf if extends to infinity)
    kaq : float, array or list
        hydraulic conductivity of each aquifer from the top down
        if float, hydraulic conductivity is the same in all aquifers
    z : array or list
        elevation tops and bottoms of the aquifers from the top down
        leaky layers may have zero thickness
        if topboundary='conf': length is 2 * number of aquifers
        if topboundary='semi': length is 2 * number of aquifers + 1 as top
        of leaky layer on top of systems needs to be specified
    c : float, array or list
        resistance of leaky layers from the top down
        if float, resistance is the same for all leaky layers
        if topboundary='conf': length is number of aquifers - 1
        if topboundary='semi': length is number of aquifers
    npor : float, array or list
        porosity of all aquifers and leaky layers from the top down
        if float, porosity is the same for all layers
        if topboundary='conf': length is 2 * number of aquifers - 1
        if topboundary='semi': length is 2 * number of aquifers
    topboundary : string, 'conf' or 'semi' (default is 'conf')
        indicating whether the top is confined ('conf') or
        semi-confined ('semi')
    hstar : float or None (default is None)
        head value above semi-confining top, only read if topboundary='semi'
    N : float (default is None)
        infiltration rate, only read if topboundary='conf'

    """
    
    def __init__(self, model, x1, x2, kaq=1, z=[1, 0], c=[], npor=0.3, 
                 topboundary='conf', hstar=None, N=None):
        self.storeinput(inspect.currentframe())
        kaq, c, npor, ltype, = param_maq(kaq, z, c, npor, topboundary)
        StripInhom.__init__(self, model, x1, x2, kaq, c, z, npor, ltype, hstar, N)


class StripInhom3D(StripInhom):
    """Create a strip inhomogeneity for a multi-layer model consisting of
    many aquifer layers. The resistance between the layers is computed
    from the vertical hydraulic conductivity of the layers.
    
    Parameters
    ----------
    model : Model object
        model to which the element is added
    x1 : float
        left boundary of inhomogeneity (may be -np.inf if extends to infinity)
    x2 : float
        right boundary of inhomogeneity (may be np.inf if extends to infinity)
    kaq : float, array or list
        hydraulic conductivity of each layer from the top down
        if float, hydraulic conductivity is the same in all aquifers
    z : array or list
        elevation of top of system followed by bottoms of all layers
        from the top down
        bottom of layer is automatically equal to top of layer below it
        if topboundary='conf': length is number of layers + 1
        if topboundary='semi': length is number of layers + 2 as top
        of leaky layer on top of systems needs to be specified
    kzoverkh : float
        vertical anisotropy ratio vertical k divided by horizontal k
        if float, value is the same for all layers
        length is number of layers
    npor : float, array or list
        porosity of all aquifer layers from the top down
        if float, porosity is the same for all layers
        if topboundary='conf': length is number of layers
        if topboundary='semi': length is number of layers + 1
    topboundary : string, 'conf' or 'semi' (default is 'conf')
        indicating whether the top is confined ('conf') or
        semi-confined ('semi')
    topres : float
        resistance of top semi-confining layer, only read if topboundary='semi'
    topthick: float
        thickness of top semi-confining layer, only read if topboundary='semi'
    hstar : float or None (default is None)
        head value above semi-confining top, only read if topboundary='semi'
    N : float (default is None)
        infiltration rate, only read if topboundary='conf'
    
    """
        
    def __init__(self, model, x1, x2, kaq, z=[1, 0], kzoverkh=1, npor=0.3, 
                 topboundary='conf', hstar=None, topres=None, topthick=0.0, N=None):
        self.storeinput(inspect.currentframe())
        kaq, c, npor, ltype, = param_3d(kaq, z, kzoverkh, npor, topboundary, topres)
        if topboundary== 'semi':
            z = np.hstack((z[0] + topthick, z))
        StripInhom.__init__(self, model, x1, x2, kaq, c, z, npor, ltype, hstar, N)
